# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.property
   :platform: Unix, Windows
   :synopsis: Represents an ontological type reference definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.mp.ontologies.core.class_ import Class
from pyesdoc.mp.ontologies.core.enum import Enum
from pyesdoc.mp.ontologies.schemas import utils



class Type(object):
    """Represents a type within an ontology.

    :ivar name: Name of type.

    """

    def __init__(self, name):
        """Instance constructor.

        :param str name: Name of type.

        """
        self.is_class = False
        self.is_complex = len(name.split('.')) > 1
        self.is_simple = not self.is_complex

        if name.startswith("linked_to"):
            self.name, qualifier = utils.parse_type(name)
        else:
            self.name, qualifier = name, None
        self.qualifier_type = Type(qualifier) if qualifier else None

        if self.is_simple:
            self.name_of_package = None
            self.name_of_type = self.name
        else:
            self.name_of_package = self.name.split('.')[0]
            self.name_of_type = self.name.split('.')[1]


    def __repr__(self):
        """Instance string representation.

        """
        return self.name


    @property
    def is_enum(self):
        """Gets flag indicating whether type represents an enumerated type.

        """
        return self.is_complex and not self.is_class



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc._search.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Exposes document search functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import json

import requests

from pyesdoc import _api_proxy as proxy
from pyesdoc import constants
from pyesdoc._publishing import retrieve



# API search endpoint.
_ENDPOINT = "/2/summary/search"

# Document type aliases.
_DOC_TYPE_ALIASES = {
    'cmip5': {
        'model': 'cim.1.software.ModelComponent',
        'experiment': 'cim.1.activity.NumericalExperiment'
    },
    'cmip6': {
        'experiment': 'cim.2.designing.NumericalExperiment',
        'machine': 'cim.2.platform.Machine',
        'mip': 'cim.2.designing.Project',
        'model': 'cim.2.science.Model',
        'performance': 'cim.2.platform.Performance',
        'project': 'cim.2.designing.Project',
        'simulation': 'cim.2.data.Simulation'
    }
}


def search(
    project,
    document_type,
    document_version=constants.DOC_VERSION_LATEST,
    institute=None,
    sub_project=None,
    model=None,
    experiment=None,
    ):
    """Exposes documentation search endpoint.

    :param str project: Project identifier, e.g. cmip6.
    :param str document_type: Document type identifier, e.g. model.
    :param str document_version: Document version filter, i.e. latest | all.
    :param str institute: Institute identifer, e.g. ipsl.
    :param str sub_project: Sub-project identifer, e.g. damip.
    :param str model: Model identifer, e.g. ipsl-cm5a-lr.
    :param str experiment: Experiment identifer, e.g. piControl.

    :returns: Search results.
    :rtype: pyesdoc._search.SearchResult

    """
    # Set criteria.
    criteria = SearchCriteria(
        project,
        document_type,
        document_version,
        institute,
        sub_project,
        model,
        experiment
        )

    # Invoke web-service.
    params = criteria.get_url_params()
    response = proxy.invoke(requests.get, _ENDPOINT, params=params)

    # Decode & return results.
    return SearchResult(json.loads(response.text))


class SearchCriteria(object):
    """Search criteria wrapper.

    """
    def __init__(self, \
        project,
        document_type,
        document_version=constants.DOC_VERSION_LATEST,
        institute=None,
        sub_project=None,
        model=None,
        experiment=None
        ):
        """Instance constructor.

        """
        self.project = project
        self.document_type = document_type
        if self.document_type[-1] in {"s", "S"}:
            self.document_type = self.document_type[0:-1]
        self.document_version = document_version
        self.institute = institute
        self.sub_project = sub_project
        self.model = model
        self.experiment = experiment


    def get_url_params(self):
        """Returns url parameters to be sent to web-service.

        """
        params = {
            k: unicode(v).strip().lower() for k, v in (
                ("document_type", self.document_type),
                ("document_version", self.document_version),
                ("experiment", self.experiment),
                ("institute", self.institute),
                ("model", self.model),
                ("project", self.project),
                ("sub_project", self.sub_project)
                ) if v is not None
        }
        try:
            params['document_type'] = _DOC_TYPE_ALIASES[params['project']][params['document_type']]
        except KeyError:
            pass

        return params


class SearchResult(object):
    """Search result wrapper.

    """
    def __init__(self, obj):
        """Instance constructor.

        """
        self.project = obj['project']
        self.total = obj['total']
        self.items = tuple([SearchResultItem(i) for i in obj['results']])


    def __repr__(self):
        """Instance default representation.

        """
        return str({
            "count": len(self)
            })


    def __len__(self):
        """Returns number of items in search results.

        """
        return len(self.items)


    def __iter__(self):
        """Instance iterator initializer.

        """
        return iter(self.items)


    def load_document(self, identifier):
        """Loads document from remote archive.

        """
        item = None

        if isinstance(identifier, int):
            assert identifier >= 0, 'Item identifier out of bounds'
            assert identifier <= len(self.items), 'Item identifier out of bounds'
            item = self.items[identifer]
        elif isinstance(identifier, basestring):
            identifier = identifier.lower()
            for i in self.items:
                if identifier in (n.lower() for n in i.names):
                    item = i

        if item:
            return item.load_document()


class SearchResultItem(object):
    """A search result item wrapper.

    """
    def __init__(self, obj):
        """Instance constructor.

        """
        self.document = None
        self.experiment = obj[0]
        self.institute = obj[1]
        self.long_name = obj[2]
        self.model = obj[3]
        self.name = obj[4]
        self.canonical_name = obj[5]
        self.uid = obj[6]
        self.version = obj[7]
        self.alternative_name = obj[8]


    def __repr__(self):
        """Instance default representation.

        """
        return str(self.__dict__)


    @property
    def names(self):
        """Returns set of names.

        """
        names = (
            self.canonical_name,
            self.name,
            self.alternative_name,
            self.long_name
            )

        return (i for i in names if i is not None and len(i) > 0)


    def load_document(self):
        """Loads document from remote archive.

        """
        if self.document is None:
            self.document = retrieve(self.uid, self.version)

        return self.document




# -*- coding: utf-8 -*-

"""
.. module:: xml.decoder.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Decodes a document from an XML text blob.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import xml.etree.ElementTree as ET

from pyesdoc import ontologies
from pyesdoc.utils import convert



# Set of supported ontology types.
_TYPES = ontologies.get_types()


def _create_doc(elem, doc_type):
    """Creates & returns a document to be hydrated from an xml element.

    """
    type_key = elem.get('ontologyTypeKey')
    if type_key and ontologies.get_type_key(doc_type) != type_key:
        doc_type = ontologies.get_type_from_key(type_key)
    if doc_type is None:
        raise ValueError('Decoding type is unrecognized')

    return doc_type(), ontologies.get_decoder_info(doc_type)


def _decode_simple(elem, typeof, is_iterable):
    """Decodes a simple type value.

    """
    if is_iterable:
        return [convert.text_to_typed_value(i.text, typeof) for i in elem]
    return convert.text_to_typed_value(elem.text, typeof)


def _decode(elem, doc_type, is_iterable):
    """Decodes a complex type value.

    """
    def _do(xml):
        # Set doc.
        doc, doc_type_info = _create_doc(xml, doc_type)

        # Set doc attributes:
        for _name, _type, _is_iterable in doc_type_info:
            # ... set xml element;
            _elem = xml.find(convert.str_to_camel_case(_name))

            # ... skip placeholders;
            if _elem is None:
                continue

            # ... complex types;
            elif _type in _TYPES:
                setattr(doc, _name, _decode(_elem, _type, _is_iterable))

            # ... simple types;
            else:
                setattr(doc, _name, _decode_simple(_elem, _type, _is_iterable))

        return doc

    return [_do(i) for i in elem] if is_iterable else _do(elem)


def decode(as_xml):
    """Decodes a document from an xml string.

    :param as_xml: Document xml representation.
    :type as_xml: unicode | str | ET.Element

    :returns: A pyesdoc document instance.
    :rtype: object

    """
    # Convert to etree.
    if isinstance(as_xml, unicode):
        as_xml = ET.fromstring(convert.unicode_to_str(as_xml))
    elif isinstance(as_xml, str):
        as_xml = ET.fromstring(as_xml)
    if not isinstance(as_xml, ET.Element):
        raise TypeError("Cannot decode non xml documents")

    # Get document type key.
    doc_type_key = as_xml.find('meta/type').text

    # Get document type.
    doc_type = ontologies.get_type_from_key(doc_type_key)
    if doc_type is None:
        raise ValueError('meta.type key cannot be mapped to an ontology type.')

    return _decode(as_xml, doc_type, False)



# -*- coding: utf-8 -*-
"""
.. module:: doc_parser.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Exposes document parsers.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.archive import config
from pyesdoc.archive.parsing import doc_parser_cmip5_qc
from pyesdoc.archive.parsing import doc_parser_cmip5_q
from pyesdoc.archive.parsing import doc_parser_esdoc_q



# Set of document parsers mapped by encoding, project & source.
_PARSERS = {
    "xml-metafor-cim-v1": {
        "cmip5": {
            "dkrz-qc": doc_parser_cmip5_qc,
            "metafor-q": doc_parser_cmip5_q
        }
    },
    "xml": {
        "esps": {
            "esdoc-q": doc_parser_esdoc_q
        },
        "downscaling-metadata": {
            "esdoc-q": doc_parser_esdoc_q
        }
    }
}


# Document type keys.
_TYPE_KEY_DOCUMENT_SET = 'cim.1.misc.documentset'



def _get_source_description(source):
    """Returns document source description.

    """
    for source_cfg in config.get_sources():
        if source_cfg.name == source:
            return source_cfg.description
    return source


def _get_children(doc):
    """Unpacks child documents contained within document being processed.

    :param object doc: Document being processed.

    :returns: Collection of documents being processed.
    :rtype: list

    """
    children = []
    if doc.type_key.lower() == _TYPE_KEY_DOCUMENT_SET:
        children += [
            doc.experiment,
            doc.model,
            doc.platform,
            doc.simulation
            ]
        children += doc.data
        children += doc.ensembles
        children += doc.grids

    return [c for c in children if c]


def _set_meta_attributes(project, source, document):
    """Sets document meta attributes.

    """
    document.meta.project = unicode(project)
    document.meta.source = _get_source_description(source)
    document.meta.source_key = unicode(source)


def _set_defaults(doc):
    """Set default institute name.

    """
    if doc.meta.institute is None:
        doc.meta.institute = u"--"


def _get_parsers(project, source):
    """Returns set of document parsers to be invoked."""
    result = ()
    mod = _get_parsing_module(project, source)
    if mod:
        try:
            result += mod.get_parsers()
        except TypeError:
            result += (mod.get_parsers(), )
    result += (_set_defaults, )

    return result


def parse(encoding, project, source, document):
    """Parses document according to project & source.

    :param encoding: Original document encoding.
    :param str project: Name of a supported project (e.g. cmip6).
    :param str source: Name of a document source (e.g. esdoc-q).
    :param object document: Document being parsed.

    """
    # Set collection of documents to process.
    docs = [document] + _get_children(document)

    # Set common meta attributes.
    for doc in docs:
        _set_meta_attributes(project, source, doc)

    # Invoke parser.
    try:
        parser = _PARSERS[encoding][project][source]
    except KeyError:
        pass
    else:
        for func in parser.PARSERS:
            for doc in docs:
                func(doc)

    # Set defaults.
    for doc in docs:
        _set_defaults(doc)



"""
A realm key-properties sepecialization.
For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Charlotte Pascoe'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Atmosphere key properties'

# --------------------------------------------------------------------
# Top level details.
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': "Top level key properties",
    'properties': [
        ('model_name', 'str', '1.1',
            'Name of atmosphere model code (CAM 4.0, ARPEGE 3.2,...)'),
        ('model_family', 'ENUM:model_family_type', '1.1',
            'Type of atmospheric model.'),
        ('basic_approximations', 'ENUM:basic_approximations_attributes', '1.N',
            'Basic approximations made in the atmosphere.',),
        ]
    }

DETAILS['resolution'] = {
    'description': "Characteristics of the model resolution",
    'properties': [
        ('horizontal_resolution_name', 'str', '1.1',
            'This is a string usually used by the modelling group to describe the resolution of the model grid, '
            'e.g. T42, N48.',),
        ('canonical_horizontal_resolution', 'str', '1.1',
            'Expression quoted for gross comparisons of resolution, e.g. 2.5 x 3.75 degrees lat-lon.'),
        ('number_of_vertical_levels', 'int', '1.1',
             "Number of vertical levels resolved on the computational grid."),
        ('high_top', 'bool', '1.1',
            "Does the atmosphere have a high-top? "
            "High-Top atmospheres have a fully resolved stratosphere with a model top above the stratopause."),
        ('timestep_dynamics', 'str', '1.1',
             "Timestep for the dynamics, e.g. 30 min."),
        ('timestep_radiative_transfer', 'str', '0.1',
             "Timestep for the radiative transfer, e.g. 3 hours."),
        ]
    }


DETAILS['orography'] = {
    'description': "Characteristics of the model orography",
    'properties': [
        ('type', 'ENUM:orography_type', '1.1',
            'Time adaptation of the orography.',),
        ('changes', 'ENUM:orography_changes', '1.N',
            'If the orography type is modified describe the time adaptation changes.'),
        ]
    }

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['model_family_type'] = {
    'description': "Type of atmospheric model",
    'is_open': True,
    'members': [
        ('AGCM', 'Atmospheric General Circulation Model'),
        ('ARCM', 'Atmospheric Regional Climate Model'),
        ]
    }

ENUMERATIONS['basic_approximations_attributes'] = {
    'description': "Basic approximations made in the atmosphere",
    'is_open': True,
    'members': [
        ('primitive equations', None),
        ('non-hydrostatic', None),
        ('anelastic', None),
        ('Boussinesq', None),
        ('hydrostatic', None),
        ('quasi-hydrostatic', None),
        ]
    }


ENUMERATIONS['orography_type'] = {
    'description': "Time adaptation of the orography",
    'is_open': False,
    'members': [
        ('present day', None),
        ('modified', None),
        ]
    }

ENUMERATIONS['orography_changes'] = {
    'description': "If the orography type is modified describe the time adaptation changes",
    'is_open': False,
    'members': [
        ('related to ice sheets', None),
        ('related to tectonics', None),
        ('modified mean', None),
        ('modified variance if taken into account in model (cf gravity waves)', None),
        ]
    }



"""A realm key-properties sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Land surface key properties'

# --------------------------------------------------------------------
# KEY PROPERTIES: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'General key properties in land surface',
    'properties': [
        ('model_name', 'str', '1.1',
             'Name of land surface model code (e.g. MOSES2.2)'),
        ('description', 'str', '1.1',
            'General description of the processes modelled (e.g. dymanic vegation, prognostic albedo, etc.)'),
        ('land_atmosphere_flux_exchanges', 'ENUM:land_atmosphere_flux_exchanges_types', '0.N',
            'Fluxes exchanged with the atmopshere.'),
        ('atmospheric_coupling_treatment', 'str', '1.1',
            'Describe the treatment of land surface coupling with the Atmosphere model component, which may be different for different quantities (e.g. dust: semi-implicit, water vapour: explicit)',),
        ('land_cover', 'ENUM:land_cover_types', '1.N',
            'Types of land cover defined in the land surface model'),
        ('land_cover_change', 'str', '0.1',
             'Describe how land cover change is managed (e.g. the use of net or gross transitions)'),
        ('tiling', 'str', '1.1',
             'Describe the general tiling procedure used in the land surface (if any). Include treatment of physiography, land/sea, (dynamic) vegetation coverage and orography/roughness'),
    ]
}

DETAILS['conservation_properties'] = {
    'description': 'TODO',
    'properties' : [
        ('energy', 'str', '0.1',
             'Describe if/how energy is conserved globally and to what level (e.g. within X [units]/year)'),
        ('water', 'str', '0.1',
             'Describe if/how water is conserved globally and to what level (e.g. within X [units]/year)'),
        ('carbon', 'str', '0.1',
             'Describe if/how carbon is conserved globally and to what level (e.g. within X [units]/year)'),
        ]
    }

DETAILS['timestepping_framework'] = {
    'description': 'TODO',
    'properties' : [
        ('timestep_dependent_on_atmosphere', 'bool', '1.1',
            'Is a time step dependent on the frequency of atmosphere coupling?'),
        ('time_step', 'int', '1.1',
            'Overall timestep of land surface model (i.e. time between calls)'),
        ('timestepping_method', 'str', '1.1',
            'General description of time stepping method and associated time step(s)'),
        ]
    }

DETAILS['software_properties'] = {
    'description': 'Software properties of land surface code',
    'properties':[
        ('repository','str', '0.1',
            "Location of code for this component."),
        ('code_version','str', '0.1',
            "Code version identifier."),
        ('code_languages','str', '0.N',
            "Code language(s)."),
    ]
}

# --------------------------------------------------------------------
# KEY PROPERTIES: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['land_atmosphere_flux_exchanges_types'] = {
    'description': 'TODO',
    'is_open': True,
    'members': [
        ('water', None),
        ('energy', None),        
        ('carbon', None),
        ('nitrogen', None),
        ('phospherous', None),
    ]
}

ENUMERATIONS['atmospheric_coupling_types'] = {
    'description': 'Treatments of land surface coupling with the Atmosphere model component',
    'is_open': True,
    'members': [
        ('implicit', None),
        ('semi-implicit', None),
        ('explicit', None),
        ]
    }

ENUMERATIONS['land_cover_types'] = {
    'description': 'Types of land cover defined in the land surface model',
    'is_open': True,
    'members': [
        ('bare soil', None),
        ('urban', None),
        ('lake', None),
        ('land ice', None),
        ('lake ice', None),
        ('vegetated', None),
        ]
    }

ENUMERATIONS['prognostic_vars_types'] = {
    'description': 'Prognostic variables included in the land surface scheme',
    'is_open': True,
    'members': [
        ('soil temperature', None),
        ('soil heat content', None),
        ('soil moisture', None),
        ('soil ice content', None),
        ('snow mass', None),
        ('snow albedo', None),
        ('snow density', None),
        ('snow water content', None),
        ('snow layer thickness', None),
        ('canopy water content', None),
        ('canopy heat content', None),
        ('canopy snow content', None),
        ('canopy skin temperature', None),
        ('canopy height', None),
        ('surface skin temperature', None),
        ('river water discharge', None),
        ('soil carbon', None),
        ('vegetation carbon', None),
        ('nitrogen', None),
        ]
    }



from pyesdoc.mp.ontologies.generators.python import utils
from pyesdoc.mp.ontologies.generators.python.decoder_generator import DecoderGenerator
from pyesdoc.mp.ontologies.generators.python.package_extended_schema_generator import PackageExtendedSchemaGenerator
from pyesdoc.mp.ontologies.generators.python.package_typeset_generator import PackageTypeSetGenerator



# Expose utility functions.
UTILS = utils

# Set of supported custom generators or tornado templates.
GENERATORS = {
    '__init__.tornado',
    'type_info.tornado',
    'typeset.tornado',
    PackageExtendedSchemaGenerator,
    PackageTypeSetGenerator,
    DecoderGenerator
}



"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe, David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Atmospheric aerosol transport'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------

# --------------------------------------------------------------------
# SUB-PROCESS: Surface emissions
# --------------------------------------------------------------------
DETAILS['surface_emissions'] = {
    'description': 'TO DO',
    'properties': [
        ('method', 'ENUM:emissions_methods', '0.N',
            'Method used to define aerosol species emitted at the surface (several methods allowed because the different species may not use the same method).'),
        ('sources', 'ENUM:surface_source_types', '0.N',
             'Sources of the aerosol species emitted at the surface that are taken into account in the emissions scheme'),
        ('prescribed_climatology', 'ENUM:prescribed_climatology_type', '0.1',
            'Specify the climatology type for aerosol emissions prescribed at the surface'),
        ('prescribed_climatology_emitted_species', 'str', '0.1',
             'List of aerosol species emitted at the surface and prescribed via a climatology'),
        ('prescribed_spatially_uniform_emitted_species', 'str', '0.1',
             'List of aerosol species emitted at the surface and prescribed as spatially uniform'),
        ('interactive_emitted_species', 'str', '0.1',
             'List of aerosol species emitted at the surface and specified via an interactive method'),
        ('other_emitted_species', 'str', '0.1',
             'List of aerosol species emitted at the surface and specified via an "other method"'),
        ('other_method_characteristics', 'str', '0.1',
             'Characteristics of the "other method" used for aerosol emissions at the surface'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Atmospheric emissions
# --------------------------------------------------------------------
DETAILS['atmospheric_emissions'] = {
    'description': 'TO DO',
    'properties': [
        ('method', 'ENUM:emissions_methods', '0.N',
            'Method used to define the aerosol species emitted in the atmosphere (several methods allowed because the different species may not use the same method).'),
        ('sources', 'ENUM:atmospheric_source_types', '0.N',
             'Sources of aerosol species emitted in the atmosphere that are taken into account in the emissions scheme.'),
        ('prescribed_climatology', 'ENUM:prescribed_climatology_type', '0.1',
            'Specify the climatology type for aerosol emissions prescribed in the atmosphere.'),
        ('prescribed_climatology_emitted_species', 'str', '0.1',
             'List of aerosol species emitted in the atmosphere and prescribed via a climatology'),
        ('prescribed_spatially_uniform_emitted_species', 'str', '0.1',
             'List of aerosol species emitted in the atmosphere and prescribed as spatially uniform'),
        ('interactive_emitted_species', 'str', '0.1',
             'List of aerosol species emitted in the atmosphere and specified via an interactive method'),
        ('other_emitted_species', 'str', '0.1',
             'List of aerosol species emitted in the atmosphere and specified via an "other method"'),
        ('other_method_characteristics', 'str', '0.1',
             'Characteristics of the "other method" used for aerosol emissions in the atmosphere'),
        
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Concentrations
# --------------------------------------------------------------------
DETAILS['concentrations'] = {
    'description': 'TO DO',
    'properties': [
        ('prescribed_lower_boundary', 'str', '0.1',
            'List of species prescribed at the lower boundary.'),
        ('prescribed_upper_boundary', 'str', '0.1',
            'List of species prescribed at the upper boundary.'),
    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['prescribed_climatology_type'] = {
    'description': 'Specify the climatology type for aerosol emissions',
    'is_open': False,
    'members': [
        ('Constant', None),
        ('Interannual', None),
        ('Annual', None),
        ('Monthly', None),
        ('Daily', None),
        ]
    }

ENUMERATIONS['surface_source_types'] = {
    'description':'Sources of the aerosol species emitted at the surface that are taken into account in the emissions scheme',
    'is_open': True,
    'members':[
        ('Vegetation', None),
        ('Bare ground', None),
        ('Sea surface', None),
        ('Anthropogenic', None),
    ]
}

ENUMERATIONS['atmospheric_source_types'] = {
    'description':'Sources of the aerosol species emitted in the atmosphere that are taken into account in the emissions scheme',
    'is_open': True,
    'members':[
        ('Aircraft', None),
        ('Biomass burning', None),
        ('Lightning', None),
        ('Volcanos', None),
    ]
}

ENUMERATIONS['emissions_methods'] = {
    'description': 'Method used to define aerosol species emitted (several methods allowed because the different species may not use the same method).',
    'is_open': True,
    'members':[
        ('Prescribed (climatology)', None),
        ('Prescribe (spatially uniform)', None),
        ('Interactive', None),
    ]
}



# -*- coding: utf-8 -*-

"""
.. module:: exceptions.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Package exceptions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import inspect



class LibraryBaseException(Exception):
    """Default library exception class.

    """

    def __init__(self, msg):
        """Contructor.

        :param msg: Exception message.
        :type msg: str

        """
        self.message = msg() if inspect.isfunction(msg) else unicode(msg)


    def __str__(self):
        """Returns a string representation.

        """
        return u"ES-DOC PY-CLIENT EXCEPTION : {0}".format(repr(self.message))


class ObsoleteException(LibraryBaseException):
    """Raised when an obsolete documents is being processed.

    """
    def __init__(self, message):
        """Instance constructor.

        """
        super(ObsoleteException, self).__init__("WARNING :: obsolete document :: {0}".format(message))


class DecodingException(LibraryBaseException):
    """Raised when document decoding fails.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(DecodingException, self).__init__("Document decoding failed.")


class EncodingException(LibraryBaseException):
    """Raised when document encoding fails.

    """
    def __init__(self, encoding, doc, error):
        """Instance constructor.

        """
        super(EncodingException, self).__init__("Document {} encoding failed :: Doc Type = {} :: Error = {}".format(encoding.upper(), type(doc), error))


class ExtendingException(LibraryBaseException):
    """Exception raised when document extension fails.

    """
    def __init__(self, err, extender=None):
        """Instance constructor.

        """
        if extender is not None:
            err = "Document extension failed: {0} :: {1}".format(extender, err)
        else:
            err = "Document extension failed: {}".format(err)
        super(ExtendingException, self).__init__(err)


class InvalidDocumentTypeException(LibraryBaseException):
    """Exception raised when an invalid document type is declared.

    """
    def __init__(self, doc_type):
        """Instance constructor.

        """
        err = "Document type [{}] is unsupported.".format(doc_type)
        super(InvalidDocumentTypeException, self).__init__(err)


class LoadingException(LibraryBaseException):
    """Exception raised when document loading fails.

    """
    def __init__(self, err):
        """Instance constructor.

        """
        err = "Document loading failed: {}".format(err)
        super(LoadingException, self).__init__(err)


class ParsingException(LibraryBaseException):
    """Exception raised when document parsing fails.

    """
    def __init__(self, err):
        """Instance constructor.

        """
        err = "Document parsing failed: {}".format(err)
        super(ParsingException, self).__init__(err)


class WebServiceException(LibraryBaseException):
    """Exception raised when a web service operation fails.

    """
    def __init__(self, err):
        """Instance constructor.

        """
        err = "Web service invocation failed: {}".format(err)
        super(ParsingException, self).__init__(err)




# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.generators.qxml.root_generator.py
   :platform: Unix, Windows
   :synopsis: Generates root package initialisation code.

.. moduleauthor:: Allyn Treshansky <allyn.treshansky@noaa.gov>


"""
from pyesdoc.mp.ontologies.generators import generator_utils as gu
from pyesdoc.mp.ontologies.generators.generator import Generator
from pyesdoc.mp.ontologies.generators.qxml import utils as qgu

from lxml import etree as et
from functools import reduce
from errno import EEXIST as PATH_EXISTS
import datetime
import os

# Generator language.
_LANG = 'qxml'


class RootGenerator(Generator):
    """Generates root level packages.

    """
    def on_ontology_parse(self, ctx):
        """Event handler for the ontology parse event.

        :param GeneratorContext ctx: Generation context information.

        """
        ontology = ctx.ontology

        name_node = et.Element("name")
        name_node.text = qgu.get_ontology_name(ontology)
        version_node = et.Element("version")
        version_node.text = qgu.get_ontology_version(ontology)
        classes_node = et.Element("classes")

        root_node = ctx.node
        root_node.append(name_node)
        root_node.append(version_node)
        root_node.append(classes_node)

    # QXML is a single construct; it is not separated into packages
    # def on_package_parse(self, ctx):
    #     pass

    def on_class_parse(self, ctx):
        """Event handler for the class parse event.

        :param GeneratorContext ctx: Generation context information.

        """
        kls = ctx.cls

        if not kls.is_abstract:  # and not kls.is_document_meta:

            class_node = et.Element("class")

            is_meta_class = kls.is_document_meta
            class_node.set("is_meta", "true" if is_meta_class else "false")

            if kls.is_entity:
                class_node.set("stereotype", "document")

            class_node.set("package", ctx.pkg.name)
            class_name_node = et.Element("name")
            class_name_node.text = kls.name
            class_node.append(class_name_node)
            class_description_node = et.Element("description")
            class_description_node.text = qgu.clean_xml_text(kls.doc_string)
            class_node.append(class_description_node)

            class_label = kls.pstr
            if class_label is not None:
                class_label_node = et.Element("label")
                class_label_text_node = et.Element("text")
                class_label_text_node.text = class_label.text
                class_label_node.append(class_label_text_node)
                class_label_fields_node = et.Element("fields")
                for class_label_field in class_label.fields:
                    class_label_field_node = et.Element("field")
                    class_label_field_node.text = class_label_field
                    class_label_fields_node.append(class_label_field_node)
                class_label_node.append(class_label_fields_node)
                class_node.append(class_label_node)

            class_attributes_node = et.Element("attributes")
            all_attributes = kls.all_properties
            # non_meta_attributes = [a for a in all_attributes if not a.is_linked_document]
            for attribute in all_attributes:
                attribute_node = et.Element("attribute")
                is_meta_property = attribute.is_linked_document
                attribute_node.set("is_meta", "true" if is_meta_property else "false")
                attribute_node.set("package", attribute.package.name)
                attribute_node.set("is_nillable", "true" if attribute.min_occurs == '0' else "false")
                attribute_name_node = et.Element("name")
                attribute_name_node.text = attribute.name
                attribute_description_node = et.Element("description")
                attribute_description_node.text = qgu.clean_xml_text(attribute.doc_string)
                attribute_cardinality_node = et.Element("cardinality")
                attribute_cardinality_node.set("min", attribute.min_occurs)
                attribute_cardinality_node.set("max", attribute.max_occurs)

                attribute_type = qgu.get_property_type(attribute)
                attribute_type_node = et.Element("type")
                attribute_type_node.text = attribute_type

                attribute_node.append(attribute_name_node)
                attribute_node.append(attribute_description_node)
                attribute_node.append(attribute_cardinality_node)
                attribute_node.append(attribute_type_node)

                if attribute_type == qgu.QXML_ATOMIC_TYPE:
                    attribute_atomic_node = et.Element("atomic")
                    attribute_atomic_type_node = et.Element("atomic_type")
                    attribute_atomic_type_node.text = qgu.get_atomic_property_type(attribute)
                    attribute_atomic_node.append(attribute_atomic_type_node)
                    attribute_node.append(attribute_atomic_node)

                elif attribute_type == qgu.QXML_ENUMERATION_TYPE:
                    attribute_enumeration_node = et.Element("enumeration")
                    attribute_enumeration_node.set("enumeration_name", attribute.type.name)
                    # TODO: WORK OUT HOW "is_open" IS SET...
                    attribute_is_open = False
                    attribute_is_multi = attribute.max_occurs != '0'
                    attribute_enumeration_node.set("is_open", "true" if attribute_is_open else "false")
                    attribute_enumeration_node.set("is_multi", "true" if attribute_is_multi else "false")
                    attribute_choices_node = et.Element("choices")
                    # attribute_node is completed (ie: the choices are added) in the 'on_enum_parse' fn below...
                    # (that is why I record the fully-qualified "enumeration_name" XML attribute above)
                    attribute_enumeration_node.append(attribute_choices_node)
                    attribute_node.append(attribute_enumeration_node)

                elif attribute_type == qgu.QXML_RELATIONSHIP_TYPE:
                    attribute_relationship_node = et.Element("relationship")
                    attribute_relationship_targets_node = et.Element("targets")
                    attribute_relationship_targets = qgu.get_relationship_property_target_classes(attribute)
                    for attribute_relationship_target in attribute_relationship_targets:
                        attribute_relationship_target_node = et.Element("target")
                        attribute_relationship_target_node.text = "{0}.{1}".format(attribute_relationship_target.package, attribute_relationship_target.name)
                        attribute_relationship_targets_node.append(attribute_relationship_target_node)
                    attribute_relationship_node.append(attribute_relationship_targets_node)
                    attribute_node.append(attribute_relationship_node)

                class_attributes_node.append(attribute_node)

            class_node.append(class_attributes_node)

            classes_node = ctx.node.xpath("//classes")[0]
            classes_node.append(class_node)

    def on_enum_parse(self, ctx):
        """Event handler for the enum parse event.

        :param GeneratorContext ctx: Generation context information.

        """
        return

        enum = ctx.enum
        enum_name = qgu.get_qualified_enum_name(enum)
        enum_choices = enum.members
        enum_xpath = "//attribute[enumeration/@enumeration_name='{0}']".format(enum_name)

        attribute_nodes = ctx.node.xpath(enum_xpath)
        for attribute_node in attribute_nodes:
            enumeration_choices_node = attribute_node.xpath("enumeration/choices")[0]

            for choice in enum_choices:
                enumeration_choice_node = et.Element("choice")
                choice_value = choice.name
                choice_description = qgu.clean_xml_text(choice.doc_string)
                enumeration_choice_value_node = et.Element("value")
                enumeration_choice_value_node.text = choice_value
                enumeration_choice_description_node = et.Element("description")
                if choice_description:
                    enumeration_choice_description_node.text = choice_description
                enumeration_choice_node.append(enumeration_choice_value_node)
                enumeration_choice_node.append(enumeration_choice_description_node)
                enumeration_choices_node.append(enumeration_choice_node)

    def on_start(self, ctx):
        """Event handler for the start parse event.

        :param GeneratorContext ctx: Generation context information.

        """

        root_node = et.Element("ontology")
        comment_node = et.Comment(" created by esdoc-mp ")
        root_node.append(comment_node)
        ctx.set_node(root_node)

    def on_end(self, ctx):
        """Event handler for the end parse event.

        :param GeneratorContext ctx: Generation context information.

        """

        output_filepath = qgu.get_ontology_path(ctx)
        if not os.path.exists(os.path.dirname(output_filepath)):
            try:
                os.makedirs(os.path.dirname(output_filepath))
            except OSError as error:  # handle race condition, as per http://stackoverflow.com/a/12517490
                if error.errno != PATH_EXISTS:
                    raise error
        with open(output_filepath, 'w') as f:
            qxml_content = et.tostring(ctx.node, pretty_print=True)
            f.write(qxml_content)
        f.closed



# -*- coding: utf-8 -*-

"""
.. module:: cim.v2.typeset_for_science_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The set of types of the cim.v2.science package.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
import abc
import datetime
import uuid

import typeset_for_shared_package as shared
import typeset_for_software_package as software



class Model(software.ComponentBase):
    """A concrete class within the cim v2 type system.

    A model component: can be executed standalone, and has as
    scientific description available via a link to a science.domain
    document. (A configured model can be understood in terms of a
    simulation, a model, and a configuration.).

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Model, self).__init__()

        self.activity_properties = None                   # science.Topic (0.1)
        self.coupled_components = []                      # science.Model (0.N)
        self.coupler = None                               # software.CouplingFramework (0.1)
        self.internal_software_components = []            # software.SoftwareComponent (0.N)
        self.key_properties = None                        # science.Topic (0.1)
        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)
        self.model_type = None                            # science.ModelTypes (1.1)
        self.realms = []                                  # science.Realm (0.N)


class Topic(object):
    """A concrete class within the cim v2 type system.

    An organized collection of details upon a specific topic, e.g. model key properties.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Topic, self).__init__()

        self.citations = []                               # shared.Citation (0.N)
        self.description = None                           # unicode (1.1)
        self.keywords = []                                # unicode (0.N)
        self.overview = None                              # unicode (0.1)
        self.properties = []                              # science.TopicProperty (0.N)
        self.property_sets = []                           # science.TopicPropertySet (0.N)
        self.responsible_parties = []                     # shared.Responsibility (0.N)
        self.short_name = None                            # unicode (1.1)
        self.specialization_id = None                     # unicode (1.1)
        self.sub_topics = []                              # science.Topic (0.N)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{}".format(self.s, self.h, self.o, self.r, self.t, self._, self.n, self.a, self.m, self.e)


class TopicProperty(object):
    """A concrete class within the cim v2 type system.

    A specialized question asked of the scientic community.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(TopicProperty, self).__init__()

        self.qc_status = None                             # int (1.1)
        self.specialization_id = None                     # unicode (1.1)
        self.values = []                                  # unicode (1.N)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{}: {}".format(self.specialization_id, self.values)


class TopicPropertySet(object):
    """A concrete class within the cim v2 type system.

    Provides specific details related to a topic (i.e. process, sub-process,
    grid, key properties, etc).

    """
    def __init__(self):
        """Instance constructor.

        """
        super(TopicPropertySet, self).__init__()

        self.description = None                           # unicode (1.1)
        self.properties = []                              # science.TopicProperty (1.N)
        self.short_name = None                            # unicode (1.1)
        self.specialization_id = None                     # unicode (1.1)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{}".format(self.specialization_id)


class Realm(Topic):
    """A concrete class within the cim v2 type system.

    Scientific area of a numerical model - usually a sub-model or component.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Realm, self).__init__()

        self.canonical_name = None                        # unicode (0.1)
        self.grid = None                                  # science.Topic (0.1)
        self.key_properties = None                        # science.Topic (0.1)
        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)
        self.model = None                                 # science.Model (1.1)
        self.processes = []                               # science.Topic (1.N)
        self.software_frameworks = []                     # software.Implementation (0.N)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{}".format(self.c, self.a, self.n, self.o, self.n, self.i, self.c, self.a, self.l, self._, self.n, self.a, self.m, self.e)


class ModelTypes(object):
    """An enumeration within the cim v2 type system.

    Defines a set of gross model classes.
    """
    is_open = False
    members = [
        "Atm Only",
        "Ocean Only",
        "Regional",
        "ESM",
        "GCM",
        "IGCM",
        "GCM-MLO",
        "Mesoscale",
        "Process",
        "Planetary"
        ]
    descriptions = [
        "Atmosphere Only",
        "Ocean Only",
        "Regional Model",
        "Earth System Model (Atmosphere, Ocean, Land, Sea-ice and cycles)",
        "Global Climate Model (Atmosphere, Ocean, no carbon cycle)",
        "Intermediate Complexity GCM",
        "GCM with mixed layer ocean",
        "Mesoscale Model",
        "Limited Area process model",
        "Non-Earth model"
        ]





# -*- coding: utf-8 -*-

"""
.. module:: json.encoder.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Encodes a document as a JSON text blob.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
from pyesdoc._codecs.dictionary import encoder as dict_encoder
from pyesdoc.utils import convert



def encode(doc):
    """Encodes a document.

    :param doc: Document being encoded.
    :type doc: object

    :returns: An encoded document representation.
    :rtype: unicode

    """
    return convert.dict_to_json(
      dict_encoder.encode(doc),
      convert.str_to_camel_case
      )




# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.__init__.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The cim v1 package initialisor.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
# Ontology name.
NAME = 'cim'

# Ontology Version.
VERSION = '1'

# Ontology full name.
FULL_NAME = '{}.{}'.format(NAME, VERSION)

# Ontology packag set.
import typeset_for_activity_package as activity
import typeset_for_data_package as data
import typeset_for_grids_package as grids
import typeset_for_misc_package as misc
import typeset_for_quality_package as quality
import typeset_for_shared_package as shared
import typeset_for_software_package as software

# Ontology class set.
from typeset_for_activity_package import Activity
from typeset_for_activity_package import BoundaryCondition
from typeset_for_activity_package import Conformance
from typeset_for_activity_package import DownscalingSimulation
from typeset_for_activity_package import Ensemble
from typeset_for_activity_package import EnsembleMember
from typeset_for_activity_package import Experiment
from typeset_for_activity_package import ExperimentRelationship
from typeset_for_activity_package import ExperimentRelationshipTarget
from typeset_for_activity_package import InitialCondition
from typeset_for_activity_package import LateralBoundaryCondition
from typeset_for_activity_package import MeasurementCampaign
from typeset_for_activity_package import NumericalActivity
from typeset_for_activity_package import NumericalExperiment
from typeset_for_activity_package import NumericalRequirement
from typeset_for_activity_package import NumericalRequirementOption
from typeset_for_activity_package import OutputRequirement
from typeset_for_activity_package import PhysicalModification
from typeset_for_activity_package import Simulation
from typeset_for_activity_package import SimulationComposite
from typeset_for_activity_package import SimulationRelationship
from typeset_for_activity_package import SimulationRelationshipTarget
from typeset_for_activity_package import SimulationRun
from typeset_for_activity_package import SpatioTemporalConstraint
from typeset_for_data_package import DataContent
from typeset_for_data_package import DataDistribution
from typeset_for_data_package import DataExtent
from typeset_for_data_package import DataExtentGeographical
from typeset_for_data_package import DataExtentTemporal
from typeset_for_data_package import DataExtentTimeInterval
from typeset_for_data_package import DataHierarchyLevel
from typeset_for_data_package import DataObject
from typeset_for_data_package import DataProperty
from typeset_for_data_package import DataRestriction
from typeset_for_data_package import DataStorage
from typeset_for_data_package import DataStorageDb
from typeset_for_data_package import DataStorageFile
from typeset_for_data_package import DataStorageIp
from typeset_for_data_package import DataTopic
from typeset_for_grids_package import CoordinateList
from typeset_for_grids_package import GridExtent
from typeset_for_grids_package import GridMosaic
from typeset_for_grids_package import GridProperty
from typeset_for_grids_package import GridSpec
from typeset_for_grids_package import GridTile
from typeset_for_grids_package import GridTileResolutionType
from typeset_for_grids_package import SimpleGridGeometry
from typeset_for_grids_package import VerticalCoordinateList
from typeset_for_misc_package import DocumentSet
from typeset_for_quality_package import CimQuality
from typeset_for_quality_package import Evaluation
from typeset_for_quality_package import Measure
from typeset_for_quality_package import Report
from typeset_for_shared_package import Calendar
from typeset_for_shared_package import Change
from typeset_for_shared_package import ChangeProperty
from typeset_for_shared_package import Citation
from typeset_for_shared_package import ClosedDateRange
from typeset_for_shared_package import Compiler
from typeset_for_shared_package import Daily360
from typeset_for_shared_package import DataSource
from typeset_for_shared_package import DateRange
from typeset_for_shared_package import DocMetaInfo
from typeset_for_shared_package import DocReference
from typeset_for_shared_package import License
from typeset_for_shared_package import Machine
from typeset_for_shared_package import MachineCompilerUnit
from typeset_for_shared_package import OpenDateRange
from typeset_for_shared_package import PerpetualPeriod
from typeset_for_shared_package import Platform
from typeset_for_shared_package import Property
from typeset_for_shared_package import RealCalendar
from typeset_for_shared_package import Relationship
from typeset_for_shared_package import ResponsibleParty
from typeset_for_shared_package import Standard
from typeset_for_shared_package import StandardName
from typeset_for_software_package import Component
from typeset_for_software_package import ComponentLanguage
from typeset_for_software_package import ComponentLanguageProperty
from typeset_for_software_package import ComponentProperty
from typeset_for_software_package import Composition
from typeset_for_software_package import Connection
from typeset_for_software_package import ConnectionEndpoint
from typeset_for_software_package import ConnectionProperty
from typeset_for_software_package import Coupling
from typeset_for_software_package import CouplingEndpoint
from typeset_for_software_package import CouplingProperty
from typeset_for_software_package import Deployment
from typeset_for_software_package import EntryPoint
from typeset_for_software_package import ModelComponent
from typeset_for_software_package import Parallelisation
from typeset_for_software_package import ProcessorComponent
from typeset_for_software_package import Rank
from typeset_for_software_package import SpatialRegridding
from typeset_for_software_package import SpatialRegriddingProperty
from typeset_for_software_package import SpatialRegriddingUserMethod
from typeset_for_software_package import StatisticalModelComponent
from typeset_for_software_package import TimeLag
from typeset_for_software_package import TimeTransformation
from typeset_for_software_package import Timing

# Ontology enum set.
from typeset_for_activity_package import ConformanceType
from typeset_for_activity_package import DownscalingType
from typeset_for_activity_package import EnsembleType
from typeset_for_activity_package import ExperimentRelationshipType
from typeset_for_activity_package import FrequencyType
from typeset_for_activity_package import ProjectType
from typeset_for_activity_package import ResolutionType
from typeset_for_activity_package import SimulationRelationshipType
from typeset_for_activity_package import SimulationType
from typeset_for_data_package import DataHierarchyType
from typeset_for_data_package import DataStatusType
from typeset_for_grids_package import ArcTypeEnum
from typeset_for_grids_package import DiscretizationEnum
from typeset_for_grids_package import FeatureTypeEnum
from typeset_for_grids_package import GeometryTypeEnum
from typeset_for_grids_package import GridNodePositionEnum
from typeset_for_grids_package import GridTypeEnum
from typeset_for_grids_package import HorizontalCsEnum
from typeset_for_grids_package import RefinementTypeEnum
from typeset_for_grids_package import VerticalCsEnum
from typeset_for_quality_package import CimFeatureType
from typeset_for_quality_package import CimResultType
from typeset_for_quality_package import CimScopeCodeType
from typeset_for_quality_package import QualityIssueType
from typeset_for_quality_package import QualitySeverityType
from typeset_for_quality_package import QualityStatusType
from typeset_for_shared_package import ChangePropertyType
from typeset_for_shared_package import CompilerType
from typeset_for_shared_package import DataPurpose
from typeset_for_shared_package import InterconnectType
from typeset_for_shared_package import MachineType
from typeset_for_shared_package import MachineVendorType
from typeset_for_shared_package import OperatingSystemType
from typeset_for_shared_package import ProcessorType
from typeset_for_shared_package import UnitType
from typeset_for_software_package import ComponentPropertyIntentType
from typeset_for_software_package import ConnectionType
from typeset_for_software_package import CouplingFrameworkType
from typeset_for_software_package import ModelComponentType
from typeset_for_software_package import SpatialRegriddingDimensionType
from typeset_for_software_package import SpatialRegriddingStandardMethodType
from typeset_for_software_package import StatisticalModelComponentType
from typeset_for_software_package import TimeMappingType
from typeset_for_software_package import TimingUnits

import type_info
import typeset
try:
	import decoder
except ImportError:
	pass



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.schema_validation.__init__.py
   :platform: Unix, Windows
   :synopsis: Validates an ontology schema definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import inspect
import re

from pyesdoc.mp.ontologies.core.schema_validation.context import ValidationContext
from pyesdoc.mp.ontologies.core.schema_validation import class_validator
from pyesdoc.mp.ontologies.core.schema_validation import enum_validator
from pyesdoc.mp.ontologies.core.schema_validation import package_validator
from pyesdoc.mp.ontologies.core.schema_validation import schema_validator
from pyesdoc.mp.ontologies.core.schema_validation import type_validator



def validate(schema):
    """Validates ontology schema.

    :param module schema: Ontology schema definition.

    :returns: Set of validation errors (if any).
    :rtype: set

    """
    ctx = ValidationContext(schema)
    for validator in (
        schema_validator,
        package_validator,
        type_validator
        ):
        validator.validate(ctx)
        if ctx.report:
            break

    return ctx.report


def is_valid(schema):
    """Returns flag indicating whether the passed schema is deemed to be valid or not.

    :param module schema: An ontology schema definition.

    :returns: True if valid false otherwise.
    :rtype: bool

    """
    return len(validate(schema)) == 0



"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Land surface energy balance'

# --------------------------------------------------------------------
# ENERGY BALANCE: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'General properties in land surface energy balance scheme',
    'properties': [
        ('tiling', 'str', '0.1',
             'Describe the energy balance tiling, if any.'),        
        ('number_of_surface_temperatures', 'int', '1.1',
            'The maximum number of distinct surface temperatures in a grid cell  (for example, each subgrid tile may have its own temperature)'),
        ('evaporation', 'ENUM:evaporation_formulation_types', '1.N',
             'Specify the formulation method for land surface evaporation, from soil and vegetation'),
        ('processes', 'ENUM:process_types', '1.N',
             'Describe which processes are included in the energy balance scheme'),
        ]
    }

# --------------------------------------------------------------------
# ENERGY BALANCE: ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['evaporation_formulation_types'] = {
    'description': 'Specify the formulation method for land surface evaporation',
    'is_open': True,
    'members': [
        ('alpha', None),
        ('beta', None),        
        ('combined', None),
        ('Monteith potential evaporation', None),
    ]
}

ENUMERATIONS['process_types'] = {
    'description': 'Processes which are included in the energy balance scheme',
    'is_open': True,
    'members': [
        ('transpiration', None),
    ]
}



# -*- coding: utf-8 -*-

"""
.. module:: cim.v2.typeset_for_platform_package.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The set of types of the cim.v2.platform package.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
import abc
import datetime
import uuid

import typeset_for_shared_package as shared



class ComputePool(object):
    """A concrete class within the cim v2 type system.

    Homogeneous pool of nodes within a computing machine.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(ComputePool, self).__init__()

        self.accelerator_type = None                      # unicode (0.1)
        self.accelerators_per_node = None                 # int (0.1)
        self.clock_cycle_concurrency = None               # int (0.1)
        self.clock_speed = None                           # float (0.1)
        self.compute_cores_per_node = None                # int (0.1)
        self.cpu_type = None                              # unicode (0.1)
        self.description = None                           # unicode (0.1)
        self.interconnect = None                          # unicode (0.1)
        self.memory_per_node = None                       # platform.StorageVolume (0.1)
        self.model_number = None                          # unicode (0.1)
        self.name = None                                  # unicode (0.1)
        self.number_of_nodes = None                       # int (0.1)
        self.operating_system = None                      # unicode (0.1)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{}".format(self.name)


    @property
    def total_cores(self):
	    """Returns a computed property.

	    """
	    return self.compute_cores_per_node * self.number_of_nodes


    @property
    def total_memory(self):
	    """Returns a computed property.

	    """
	    return self.memory_per_node * self.number_of_nodes


class Partition(object):
    """A concrete class within the cim v2 type system.

    A major partition (component) of a computing system (aka machine).

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Partition, self).__init__()

        self.compute_pools = []                           # platform.ComputePool (1.N)
        self.description = None                           # unicode (0.1)
        self.institution = None                           # shared.Party (1.1)
        self.model_number = None                          # unicode (0.1)
        self.name = None                                  # unicode (1.1)
        self.online_documentation = []                    # shared.OnlineResource (0.N)
        self.partition = []                               # platform.Partition (0.N)
        self.storage_pools = []                           # platform.StoragePool (0.N)
        self.vendor = None                                # shared.Party (0.1)
        self.when_used = None                             # time.TimePeriod (0.1)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{}".format(self.name)


class Performance(object):
    """A concrete class within the cim v2 type system.

    Describes the properties of a performance of a configured model on
a particular system/machine.

Based on "CPMIP: Measurements of Real Computational Performance of
Earth System Models" (Balaji et. al. 2016, doi:10.5194/gmd-2016-197,
http://www.geosci-model-dev-discuss.net/gmd-2016-197/)

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Performance, self).__init__()

        self.actual_simulated_years_per_day = None        # float (0.1)
        self.compiler = None                              # unicode (0.1)
        self.complexity = None                            # int (0.1)
        self.core_hours_per_simulated_year = None         # float (0.1)
        self.coupling_cost = None                         # float (0.1)
        self.data_intensity = None                        # float (0.1)
        self.data_output_cost = None                      # float (0.1)
        self.joules_per_simulated_year = None             # float (0.1)
        self.memory_bloat = None                          # float (0.1)
        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)
        self.model = None                                 # science.Model (1.1)
        self.name = None                                  # unicode (0.1)
        self.parallelization = None                       # float (0.1)
        self.platform = None                              # platform.Machine (1.1)
        self.resolution = None                            # int (0.1)
        self.simulated_years_per_day = None               # float (0.1)
        self.subcomponent_performance = []                # platform.ComponentPerformance (0.N)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{} (sypd:{})".format(self.name, self.simulated_years_per_day)


class StoragePool(object):
    """A concrete class within the cim v2 type system.

    Homogeneous storage pool on a computing machine.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(StoragePool, self).__init__()

        self.description = None                           # unicode (0.1)
        self.name = None                                  # unicode (1.1)
        self.type = None                                  # platform.StorageSystems (0.1)
        self.vendor = None                                # shared.Party (0.1)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{}".format(self.name)


class StorageVolume(object):
    """A concrete class within the cim v2 type system.

    Platform storage volume and units.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(StorageVolume, self).__init__()

        self.units = None                                 # platform.VolumeUnits (1.1)
        self.volume = None                                # int (1.1)


    @property
    def __str__(self):
	    """Instrance string representation.

	    """
	    return "{} {}".format(self.volume, self.units)


class ComponentPerformance(Performance):
    """A concrete class within the cim v2 type system.

    Describes the simulation rate of a model component.

Based on "CPMIP: Measurements of Real Computational Performance of
Earth System Models" (Balaji et. al. 2016, doi:10.5194/gmd-2016-197,
http://www.geosci-model-dev-discuss.net/gmd-2016-197/)

    """
    def __init__(self):
        """Instance constructor.

        """
        super(ComponentPerformance, self).__init__()

        self.component = None                             # software.SoftwareComponent (1.1)


class Machine(Partition):
    """A concrete class within the cim v2 type system.

    A computer/system/platform/machine which is used for simulation.

    """
    def __init__(self):
        """Instance constructor.

        """
        super(Machine, self).__init__()

        self.meta = shared.DocMetaInfo()                  # shared.DocMetaInfo (1.1)


class StorageSystems(object):
    """An enumeration within the cim v2 type system.

    Controlled vocabulary for storage  types (including filesystems).
    """
    is_open = False
    members = [
        "Lustre",
        "GPFS",
        "isilon",
        "NFS",
        "S3",
        "PanFS",
        "Other Disk",
        "Tape - MARS",
        "Tape - MASS",
        "Tape - Castor",
        "Tape - Other"
        ]
    descriptions = [
        "Lustre parallel file system",
        "IBM GPFS (also known as IBM Spectral Scale",
        "The EMC scaleout NAS solution",
        "Generic Network File System",
        "Object file system exposing the AWS S3 interface",
        "Panasas Parallel File system",
        "Other disk based file system",
        "Tape storage system using ECMWF MARS",
        "Tape storage system using Met Office MASS",
        "Tape storage sytsem using CERN Castor",
        "Other tape based system"
        ]


class VolumeUnits(object):
    """An enumeration within the cim v2 type system.

    Appropriate storage volume units.
    """
    is_open = False
    members = [
        "GB",
        "TB",
        "PB",
        "EB",
        "TiB",
        "PiB",
        "EiB"
        ]
    descriptions = [
        "Gigabytes (1000^3)",
        "Terabytes (1000^4)",
        "Petabytes (1000^5)",
        "Exabytes (1000^6)",
        "Tebibytes (1024^4)",
        "Pebibytes (1024^5)",
        "Exbibytes (1024^6)"
        ]






import os

from pyesdoc.mp.specializations import utils_loader as loader
from pyesdoc.mp.specializations import utils_factory as factory
from pyesdoc.mp.specializations.utils_factory import CACHE



def get_model_specialization(mip_era, topic):
	"""Returns model specializations.

	:param str mip_era: MIP era that model documentation is related to.
	:param str topic: Specialziation topic.

	:result: Topic specializations.
	:rtype: TopicSpecialization

	"""
	load_model_specialization(mip_era, topic)

	return CACHE[_get_cache_key(mip_era, topic)]


def load_model_specialization(mip_era, topic):
	"""Loads model specializations into memory.

	"""
	if _get_cache_key(mip_era, topic) in CACHE:
		return

	mip_era = unicode(mip_era).strip().lower()
	topic = unicode(topic).strip().lower()
	dpath = os.path.join(os.path.dirname(__file__), mip_era)
	modules = loader.get_modules(dpath, topic)
	factory.get_specialization(modules)


def get_property_specialization(specialization_id):
	"""Returns property specializations.

	:param str specialization_id: Specialziation identifier.

	:result: Property specialization.
	:rtype: PropertySpecialization

	"""
	if len(specialization_id) == 0:
		return

	# Ensure topic is loaded.
	parts = specialization_id.split('.')
	load_model_specialization(parts[0], parts[1])

	return CACHE[specialization_id]


def _get_cache_key(mip_era, topic):
	"""Returns model specialization cache key.

	"""
	mip_era = unicode(mip_era).strip().lower()
	topic = unicode(topic).strip().lower()

	return "{}.{}".format(mip_era, topic)



"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Land surface vegetation'

# --------------------------------------------------------------------
# PROCESS: top level properties
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Key properties of the land surface vegetation',
    'properties': [
        ('description', 'str', '1.1',
             'General description of vegetation scheme'),
        ('time_step', 'int', '1.1',
             'Time step of vegetation scheme in seconds'),                
        ('dynamic_vegetation', 'bool', '1.1',
             'Is there dynamic evolution of vegetation?'),
        ('tiling', 'str', '0.1',
             'Describe the vegetation tiling, if any.'),
        ('vegetation_representation', 'ENUM:vegetation_representation_types', '1.1',
             'Vegetation classification used'),
        ('vegetation_types', 'ENUM:vegetation_types', '0.N',
             'List of vegetation types in the classification, if any'),
        ('biome_types', 'ENUM:biome_types', '0.N',
             'List of biome types in the classification, if any'),
        ('vegetation_time_variation', 'ENUM:vegetation_time_variation', '1.1',
             'How the vegetation fractions in each tile are varying with time'),
        ('vegetation_map', 'str', '0.1',
             'If vegetation fractions are not dynamically updated , describe the vegetation map used (common name and reference, if possible)'),
        ('interception', 'bool', '1.1',
             'Is vegetation interception of rainwater represented?'),
        ('phenology', 'ENUM:phenology_methods', '1.1',
             'Treatment of vegetation phenology'),
        ('phenology_description', 'str', '0.1',
             'General description of the treatment of vegetation phenology'),
        ('leaf_area_index', 'ENUM:leaf_area_index_methods', '1.1',
             'Treatment of vegetation leaf area index'),
        ('leaf_area_index_description', 'str', '0.1',
             'General description of the treatment of leaf area index'),
        ('biomass', 'ENUM:biomass_methods', '1.1',
             'Treatment of vegetation biomass '),
        ('biomass_description', 'str', '0.1',
             'General description of the treatment of vegetation biomass'),
        ('biogeography', 'ENUM:biogeography_methods', '1.1',
             'Treatment of vegetation biogeography'),
        ('biogeography_description', 'str', '0.1',
             'General description of the treatment of vegetation biogeography'),
        ('stomatal_resistance', 'ENUM:stomatal_resistance_methods', '1.N',
             'Specify what the vegetation stomatal resistance depends on'),
        ('stomatal_resistance_description', 'str', '0.1',
             'General description of the treatment of vegetation stomatal resistance'),
        ('prognostic_variables', 'str', '1.1',
             'List the prognostic variables of the vegetation scheme'),

    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['phenology_methods'] = {
    'description': 'Treatment of vegetation phenology',
    'is_open': True,
    'members': [
        ('prognostic', None),
        ('diagnostic (vegetation map)', None),
        ]
    }

ENUMERATIONS['leaf_area_index_methods'] = {
    'description': 'Treatment of vegetation leaf area index',
    'is_open': True,
    'members': [
        ('prescribed', None),
        ('prognostic', None),
        ('diagnostic', None),
        ]
    }

ENUMERATIONS['biomass_methods'] = {
    'description': 'Treatment of vegetation biomass',
    'is_open': True,
    'members': [
        ('prognostic', None),
        ('diagnostic', None),
        ]
    }

ENUMERATIONS['biogeography_methods'] = {
    'description': 'Treatment of vegetation biogeography',
    'is_open': True,
    'members': [
        ('prognostic', None),
        ('diagnostic', None),
        ]
    }

ENUMERATIONS['stomatal_resistance_methods'] = {
    'description': 'Dependancies of vegetation stomatal resistance',
    'is_open': True,
    'members': [
        ('light', None),
        ('temperature', None),
        ('water availability', None),
        ('CO2', None),
        ('O3', None),
        ]
    }

ENUMERATIONS['biome_types'] = {
    'description': 'Biome type in the classification',
    'is_open': True,
    'members': [
        ('evergreen needleleaf forest', None),
        ('evergreen broadleaf forest', None),
        ('deciduous needleleaf forest', None),
        ('deciduous broadleaf forest', None),
        ('mixed forest', None),
        ('woodland', None),
        ('wooded grassland', None),
        ('closed shrubland', None),
        ('opne shrubland', None),
        ('grassland', None),
        ('cropland', None),
        ('wetlands', None),
        ]
    }

ENUMERATIONS['vegetation_types'] = {
    'description': 'Vegetation type in the classification',
    'is_open': True,
    'members': [
        ('broadleaf tree', None),
        ('needleleaf tree', None),
        ('C3 grass', None),
        ('C4 grass', None),
        ('vegetated', None),
        ]
    }

ENUMERATIONS['vegetation_representation_types'] = {
    'description': 'Vegetation classification used',
    'is_open': True,
    'members': [
        ('vegetation types', None),
        ('biome types', None),
        ]
    }

ENUMERATIONS['vegetation_time_variation'] = {
    'description': 'How the vegetation fraction in each tile are varying with time',
    'is_open': True,
    'members': [
        ('fixed (not varying)', None),
        ('prescribed (varying from files)', None),
        ('dynamical (varying from simulation)', None),
    ]
}



"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe, David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Aerosol transport'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Top level aerosol transport properties',
    'properties': [
        ('scheme', 'ENUM:transport_schemes', '1.1',
            'Method for aerosol transport modeling'),
        ('mass_conservation_scheme', 'ENUM:mass_conservation_methods', '1.N',
            'Method used to ensure mass conservation.'),
        ('convention_scheme', 'ENUM:convection_types', '1.N',
            'Transport by convention'),
        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: Turbulence
# --------------------------------------------------------------------
DETAILS['turbulence'] = {
    'description': 'The scheme used for parametrization of the turbulent transport of aerosols',
    'properties': [
        ('method', 'ENUM:turbulence_methods', '1.1',
            'Method used for turbulence parametrization'),
        ('specific_scheme', 'str', '0.1',
             'Description of the scheme used for parametrization of the turbulent transport of aerosols, if not that used by the atmospheric chemistry model'),        
    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['transport_schemes'] = {
    'description': 'Method for aerosol transport modeling',
    'is_open': False,
    'members': [
        ('Uses Atmospheric chemistry transport scheme', None),
        ('Specific transport scheme (eulerian)', None),
        ('Specific transport scheme (semi-lagrangian)', None),
        ('Specific transport scheme (eulerian and semi-lagrangian)', None),
        ('Specific transport scheme (lagrangian)', None),
    ]
}

ENUMERATIONS['mass_conservation_methods'] = {
    'description':'Method used to ensure mass conservation',
    'is_open': True,
    'members':[
        ('Uses Atmospheric chemistry transport scheme', None),
        ('Mass adjustment', None),
        ('Concentrations positivity', None),
        ('Gradients monotonicity', None),
    ]
}

ENUMERATIONS['convection_types'] = {
    'description':'Transport by convection',
    'is_open': True,
    'members':[
        ('Uses Atmospheric chemistry transport scheme', None),
        ('Convective fluxes connected to tracers', None),
        ('Vertical velocities connected to tracers', None),
    ]
}

ENUMERATIONS['turbulence_methods'] = {
    'description':'Method used for turbulence parametrization',
    'is_open': True,
    'members':[
        ('Uses Atmospheric chemistry turbulence scheme', None),
        ('Specific turbulence scheme', None),
    ]
}



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.generators.python.typeset_generator.py
   :platform: Unix, Windows
   :synopsis: Generates code encapsulating an ontology's typeset.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.mp.ontologies.generators.python import utils as pgu
from pyesdoc.mp.ontologies.generators.generator import Generator



class PackageExtendedSchemaGenerator(Generator):
    """Generates code to represent an ontology as a set of types.

    """
    def on_package_parse(self, ctx):
        """Event handler for the package parse event.

        :param GeneratorContext ctx: Generation context information.

        """
        ctx.code.append(
            (
                ctx.get_code("extended_schema_for_package.tornado", pgu),
                pgu.get_ontology_directory(ctx),
                pgu.get_package_module_file_name(ctx.pkg, 'extended_schema')
            )
        )



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc._api_proxy.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Proxies documentaiton web-service calls.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import os
import urllib

import requests

from pyesdoc import exceptions



# HTTP response codes.
HTTP_RESPONSE_STATUS_200 = 200
HTTP_RESPONSE_STATUS_404 = 404
HTTP_RESPONSE_STATUS_500 = 500

# Web service base url.
_BASE_URL = os.getenv("ESDOC_API", r"https://api.es-doc.org")



def invoke(verb, endpoint, params=None, data=None):
    """Invokes endpoint.

    """
    # Set endpoint url.
    url = "{}{}".format(_BASE_URL, endpoint)
    if params is not None:
    	url += "?{}".format(urllib.urlencode(params))

    # Set request headers.
    headers = None if not data else {'content-type': 'application/json'}

    # Make remote call.
    try:
        return verb(url, data=data, headers=headers)
    except requests.ConnectionError:
        _throw_connection_error()
    except requests.HTTPError:
        _throw_http_error()
    except requests.Timeout:
        _throw_timeout_error()


def parse_response(resp, error_on_404=True):
    """Parses response returned from API.

    """
    if resp.status_code == HTTP_RESPONSE_STATUS_404:
        if error_on_404:
            _throw_not_found_error()
    elif resp.status_code == HTTP_RESPONSE_STATUS_500:
        _throw_server_error(resp)
    elif resp.status_code != HTTP_RESPONSE_STATUS_200:
        _throw_uncontrolled_server_error(resp)


def _throw_connection_error():
    """Throws an error.

    """
    raise exceptions.WebServiceException("Could not connect to remote API server.")


def _throw_not_found_error():
    """Throws an error.

    """
    raise exceptions.WebServiceException("Could not find remote resource.")


def _throw_http_error():
    """Throws an error.

    """
    raise exceptions.WebServiceException("Invalid HTTP response from remote API server.")


def _throw_timeout_error():
    """Throws an error.

    """
    raise exceptions.WebServiceException("Remote API server connection timed out.")


def _throw_server_error(resp):
    """Throws an error.

    """
    raise exceptions.WebServiceException("A server side failure has occurred: {0}".format(resp.text))


def _throw_uncontrolled_server_error(resp):
    """Throws an unknown server error.

    """
    raise exceptions.WebServiceException("An uncontrolled server side failure has occurred: {0}".format(resp.text))



"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ice sheet and ice shelf'

# --------------------------------------------------------------------
# KEY PROPERTIES: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'General key properties in ice sheets and ice shelves',
    'properties': [
        ('adaptive_grid', 'bool', '1.1',
             'Is an adative grid being used?'),    
        ('resolution_limit', 'float', '0.1',
             'If an adative grid is being used, what is the limit of the resolution (in metres)'),    
        ('grounding_line_method', 'ENUM:grounding_line_methods', '1.1',
            'Specify the technique used for modelling the grounding line in the ice sheet-ice shelf coupling'),
        ('ice_sheet', 'bool', '1.1',
            'Are ice sheets simulated?'),
        ('ice_shelf', 'bool', '1.1',
            'Are ice shelves simulated?'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: Mass balance
# --------------------------------------------------------------------
DETAILS['basal_mass_balance'] = {
    'description': 'TODO',
    'properties': [],
    'detail_sets': [
        'basal',
        'frontal'
    ],
}

DETAILS['mass_balance:basal'] = {
    'description': 'TODO',
    'properties': [
        ('bedrock', 'str', '0.1',
             'Describe the implementation of basal melting over bedrock'),
        ('ocean', 'str', '0.1',
             'Describe the implementation of basal melting over the ocean'),
    ]
}

DETAILS['mass_balance:frontal'] = {
    'description': 'TODO',
    'properties': [
        ('calving', 'str', '0.1',
             'Describe the implementation of calving from the front of the ice shelf'),
        ('melting', 'str', '0.1',
             'Describe the implementation of melting from the front of the ice shelf'),
    ]
}

# --------------------------------------------------------------------
# SUB-PROCESS: Dynamics
# --------------------------------------------------------------------
DETAILS['dynamics'] = {
    'description': '',
    'properties': [
        ('coupling_wth_atmosphere', 'str', '0.1',
             'Describe the coupling method between the ice sheet and ice shelf and atmosphere'),
        ('coupling_wth_ocean', 'str', '0.1',
             'Describe the coupling method between the ice sheet and ice shelf and the ocean'),
    ],
    'detail_sets': [
        'numerics',
    ]
}

DETAILS['dynamics:numerics'] = {
    'description': 'TODO',
    'properties': [
        ('timestep', 'int', '1.1',
             'Timestep (in seconds) of the ice scheme',),        
        ('approximation', 'ENUM:approximation_types', '1.N',
            'Approximation type used in modelling ice dynamics'),
    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['grounding_line_methods'] = {
    'description': 'Specify the technique used for modelling the grounding line in the ice sheet-ice shelf coupling',
    'is_open': True,
    'members': [
        ('grounding line prescribed', None),
        ('flux prescribed (Schoof)', None),
        ('fixed grid size', None),
        ('moving grid', None),
    ]
}

ENUMERATIONS['ablation_calculation_types'] = {
    'description': 'Type of scheme used to model ice sheet ablation',
    'is_open': True,
    'members': [
        ('Energy balance model', None),
        ('PDD', 'positive degree days'),
    ]
}

ENUMERATIONS['approximation_types'] = {
    'description': 'Approximation type used in modelling ice sheet dynamics',
    'is_open': True,
    'members': [
        ('SIA', None),
        ('SAA', None),
        ('full stokes', None),
    ]
}

ENUMERATIONS['ice_sheet_snow_processes'] = {
    'description': 'Processes describing snow on ice sheets',
    'is_open': True,
    'members': [
        ('snow melting', None),
        ('snow refreezing', None),
        ('blowing snow', None),
    ]
}



# -*- coding: utf-8 -*-

"""
.. module:: test_type_cim_v1_data_data_object_2.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Tests a cim.v1.DataObject instance.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
import arrow

import pyesdoc.ontologies.cim as cim
import test_utils as tu



# Test type.
DOC_TYPE = cim.v1.DataObject

# Test type display name.
DOC_TYPE_DISPLAY_NAME = "Data Object"

# Test document type.
DOC_TYPE_KEY = DOC_TYPE.type_key + '-2'

# Test representation file.
DOC_FILE = 'cim.1.data.DataObject-2.xml-metafor-cim-v1'

# Test document uid.
DOC_UID = '834151a4-978d-4627-954e-285916bb907a'

# Test document version.
DOC_VERSION = '1'

# Test document creation date.
DOC_DATE = arrow.get('2011-09-28 16:08:41').datetime

# Test document project.
DOC_PROJECT = "CMIP5"

# Test document project.
DOC_INSTITUTE = "MOHC"

# Test document author.
DOC_AUTHOR = "Metafor Questionnaire"


def _assert_doc_core(doc, is_update):
    """Assert core information."""
    tu.assert_iter(doc.citations, 2, cim.v1.Citation)


def _assert_doc_content(doc, is_update):
    """Assert content information."""
    tu.assert_iter(doc.content, 1, cim.v1.DataContent)
    c = doc.content[0]
    tu.assert_str(c.aggregation, "sum")
    tu.assert_str(c.frequency, "Other")


def _assert_doc_data_properties(doc, is_update):
    """Assert properties."""
    tu.assert_iter(doc.properties, 2, cim.v1.DataProperty)
    p = doc.properties[0]
    tu.assert_str(p.name, "TestProperty1")
    tu.assert_str(p.value, "1")


def _assert_doc_distribution(doc, is_update):
    """Assert distribution information."""
    tu.assert_object(doc.distribution, cim.v1.DataDistribution)
    tu.assert_str(doc.distribution.access, "OnlineFileHTTP")
    tu.assert_str(doc.distribution.format, "NetCDF")
    tu.assert_object(doc.distribution.responsible_party, cim.v1.ResponsibleParty)


def _assert_doc_extent(doc, is_update):
    """Assert extent information."""
    tu.assert_object(doc.extent, cim.v1.DataExtent)
    e = doc.extent
    tu.assert_object(e.geographical, cim.v1.DataExtentGeographical)
    tu.assert_float(e.geographical.east, 360.0)
    tu.assert_float(e.geographical.north, 90.0)
    tu.assert_float(e.geographical.south, -90.0)
    tu.assert_object(e.temporal, cim.v1.DataExtentTemporal)
    tu.assert_date(e.temporal.begin, "1859-12-01 00:00:00")
    tu.assert_date(e.temporal.end, "1999-12-30 00:00:00")
    tu.assert_object(e.temporal.time_interval, cim.v1.DataExtentTimeInterval)
    ti = e.temporal.time_interval
    tu.assert_int(ti.factor, -1)
    tu.assert_int(ti.radix, 50430)
    tu.assert_str(ti.unit, "day")


def _assert_doc_hierarchy_level(doc, is_update):
    """Assert hierarchy level information."""
    tu.assert_object(doc.hierarchy_level, cim.v1.DataHierarchyLevel)
    h = doc.hierarchy_level
    tu.assert_bool(h.is_open, True)
    tu.assert_str(h.name, "experiment")
    tu.assert_str(h.value, "HADGEM2_20C3M_1_D0_hus700")


def assert_doc(doc, is_update=False):
    """Asserts a document.

    :param object doc: Document being tested.
    :param bool is_update: Flag indicating whether document has been updated.

    """
    for assertor in (
        _assert_doc_core,
        _assert_doc_content,
        _assert_doc_data_properties,
        _assert_doc_extent,
        _assert_doc_hierarchy_level,
        ):
        assertor(doc, is_update)


def assert_doc1(doc, meta, ext):
    """Asserts a document.

    :param object doc: Document being tested.
    :param object meta: Document meta information.
    :param object ext: Document extension information.

    """
    return
    assert doc.child_object == []
    for i in range(1):
        c = doc.citations[i]
        assert c is not None
        assert isinstance(c, cim.v1.Citation)
        assert c.title.startswith(str(i + 1) + ' - ')
        assert c.date == arrow.get('2009-02-11').datetime
    assert len(doc.content) == 1
    assert doc.content[0].aggregation == 'sum'
    assert doc.content[0].frequency == 'Other'
    assert doc.content[0].topic.name == 'specific_humidity'
    assert doc.content[0].topic.description.find('specific" means per unit mass') >= 0
    assert doc.content[0].topic.unit == 'm s-1'
    for i in range(1):
        dp = doc.data_property[i]
        assert isinstance(dp, cim.v1.DataProperty)
        assert dp.name == 'TestProperty' + str(i + 1)
        assert dp.value == str(i + 1)
    assert doc.data_status == 'complete'
    assert doc.description.startswith('This dataset represents daily output: instantaneous daily')
    assert doc.distribution.access == 'OnlineFileHTTP'
    assert doc.distribution.format == 'NetCDF'
    assert doc.distribution.responsible_party.individual_name == 'Met Office (HC)'
    assert doc.distribution.responsible_party.organisation_name == 'Hadley Centre'
    assert doc.distribution.responsible_party.role == 'distributor'
    assert doc.extent.geographical.east == float(360)
    assert doc.extent.geographical.south == float(-90)
    assert doc.extent.geographical.west == float(0)
    assert doc.extent.geographical.north == float(90)
    assert doc.extent.temporal.begin == arrow.get('1859-12-1').datetime
    assert doc.extent.temporal.end == arrow.get('1999-12-30').datetime
    assert doc.extent.temporal.time_interval.factor == int(-1)
    assert doc.extent.temporal.time_interval.radix == int(50430)
    assert doc.extent.temporal.time_interval.unit == 'day'
    # TODO set type correctly
    # assert obj.geometry_model is None
    assert doc.hierarchy_level.name == 'experiment'
    assert doc.hierarchy_level.value == 'HADGEM2_20C3M_1_D0_hus700'
    assert doc.hierarchy_level.is_open == True
    assert doc.keyword == 'Test keyword'
    assert doc.parent_object is None
    assert doc.parent_object_reference is None
    assert doc.purpose == 'test'
    assert len(doc.restriction) == 0
    # TODO set type correctly
    # assert obj.source_simulation is None
    assert len(doc.storage) == 0


def update_doc(doc):
    """Update a document prior to republishing.

    :param object doc: Document being republished.

    """
    pass


def assert_doc_updates(doc):
    """Asserts a document after being updated.

    :param object doc: Document being tested.

    """
    pass


# -*- coding: utf-8 -*-

"""
.. module:: utils_parser.py
   :platform: Unix, Windows
   :synopsis: An event style specializations parser.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from utils_constants import *



class SpecializationParser(object):
    """An event driven CMIP6 specializations parser.

    """
    def __init__(self, project, root, short_tables):
        """Instance constructor.

        """
        self.project = project
        self.root = root
        self.short_tables = short_tables


    def run(self):
        """Runs the parser raising events as it does so.

        """
        self.on_root_parse(self.root)
        self._parse_topic(self.root)

        for st in self.root.sub_topics:
            getattr(self, "on_{}_parse".format(st.type_key))(st)
            self._parse_topic(st)
            getattr(self, "on_{}_parsed".format(st.type_key))(st)

        if self.short_tables:
            self.on_short_tables_parse(self.short_tables)
            for short_table in self.short_tables:
                self.on_short_table_parse(short_table)
                self.on_short_table_parsed(short_table)
            self.on_short_tables_parsed(self.short_tables)

        self.on_root_parsed(self.root)


    def _parse_topic(self, topic):
        """Parses a topic.

        """
        self._parse_topic_properties(topic.properties)

        for ps in topic.property_sets:
            self.on_property_set_parse(ps)
            self._parse_topic_properties(ps.properties)
            self.on_property_set_parsed(ps)

        for sp in topic[TYPE_KEY_SUBPROCESS]:
            self.on_subprocess_parse(sp)
            self._parse_topic(sp)
            self.on_subprocess_parsed(sp)


    def _parse_topic_properties(self, properties):
        """Parses a collection of properties associated with either a topic or a property-set.

        """
        for p in properties:
            self.on_property_parse(p)
            if p.enum:
                self.on_enum_parse(p.enum)
                for ec in p.enum.choices:
                    self.on_enum_choice_parse(ec)
                    self.on_enum_choice_parsed(ec)
                self.on_enum_parsed(p.enum)
            self.on_property_parsed(p)


    def on_root_parse(self, root):
        """On root parse event handler.

        """
        pass


    def on_root_parsed(self, root):
        """On root parsed event handler.

        """
        pass


    def on_grid_parse(self, grid):
        """On grid parse event handler.

        """
        pass


    def on_grid_parsed(self, grid):
        """On grid parsed event handler.

        """
        pass


    def on_keyprops_parse(self, key_props):
        """On key-properties parse event handler.

        """
        pass


    def on_keyprops_parsed(self, key_props):
        """On key-properties parsed event handler.

        """
        pass


    def on_process_parse(self, process):
        """On process parse event handler.

        """
        pass


    def on_process_parsed(self, process):
        """On process parsed event handler.

        """
        pass


    def on_subprocess_parse(self, subprocess):
        """On sub-process parse event handler.

        """
        pass


    def on_subprocess_parsed(self, subprocess):
        """On sub-process parsed event handler.

        """
        pass


    def on_property_set_parse(self, prop_set):
        """On property set parse event handler.

        """
        pass


    def on_property_set_parsed(self, prop_set):
        """On property set parsed event handler.

        """
        pass


    def on_property_parse(self, prop):
        """On property parse event handler.

        """
        pass


    def on_property_parsed(self, prop):
        """On property parsed event handler.

        """
        pass


    def on_enum_parse(self, enum):
        """On enum parse event handler.

        """
        pass


    def on_enum_parsed(self, enum):
        """On enum parsed event handler.

        """
        pass


    def on_enum_choice_parse(self, choice):
        """On enum choice parse event handler.

        """
        pass


    def on_enum_choice_parsed(self, choice):
        """On enum choice parsed event handler.

        """
        pass


    def on_short_tables_parse(self, short_tables):
        """On short tables parse event handler.

        """
        pass


    def on_short_tables_parsed(self, short_tables):
        """On short tables parsed event handler.

        """
        pass


    def on_short_table_parse(self, short_table):
        """On short table parse event handler.

        """
        pass


    def on_short_table_parsed(self, short_table):
        """On short table parsed event handler.

        """
        pass



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc._io.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Exposes document io functions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>

"""
import collections
import glob
import json
import os

import pyesdoc
from pyesdoc import constants
from pyesdoc import ontologies
from pyesdoc.utils.convert import str_to_unicode



def get_filename(doc, encoding=None):
    """Returns document file name.

    :param object document: A pyesdoc document instance.
    :param str encoding: Document encoding.

    :returns: Document file name.
    :rtype: str

    """
    if encoding is None:
        encoding = constants.ENCODING_JSON

    return '{0}_{1}_{2}.{3}'.format(doc.__class__.type_key,
                                    doc.meta.id,
                                    doc.meta.version,
                                    encoding)


def _get_filepath(io_dir, doc, encoding):
    """Returns document file path.

    """
    return os.path.join(io_dir, get_filename(doc, encoding))


def write(document, fpath, encoding=constants.ENCODING_JSON):
    """Writes a document to the file system.

    :param object document: A pyesdoc document instance.
    :param str fpath: Path to I/O directory or file.
    :param str encoding: Document encoding.

    :returns: Path to created file.
    :rtype: str

    """
    if os.path.isdir(fpath):
        fpath = _get_filepath(fpath, document, encoding)
    fpath = os.path.expanduser(fpath)
    with open(fpath, 'w') as fstream:
        fstream.write(pyesdoc.encode(document, encoding))

    return fpath


def read(fpath, encoding=None, decode=True):
    """Reads a document from file system.

    :param str fpath: Path to previously saved file.
    :param str encoding: Encoding to use during deserialization.
    :param bool decode: Flag indicating whether document will be decoded.

    :returns: A pyesdoc document instance.
    :rtype: object

    """
    fpath = os.path.expanduser(fpath)

    # Validate file path.
    if not os.path.isfile(fpath):
        raise IOError("Document file path does not exist")

    # Optionally derive encoding from file extension.
    if encoding is None:
        encoding = os.path.splitext(fpath)[1][1:]

    # Set raw content.
    with open(fpath, 'r') as fstream:
        fcontent = str_to_unicode(fstream.read())

    # Decode upon request.
    return pyesdoc.decode(fcontent, encoding) if decode else fcontent


def seek(io_dir, filter=None, latest=False):
    """Reads a document from file system.

    :param str io_dir: Path to a directory within documents exist.
    :param object filter: Filter to apply when seeking.

    :returns: A pyesdoc document instance.
    :rtype: object

    """
    if not os.path.isdir(io_dir):
        raise ValueError("I/O directory is invalid")

    # Set search targets.
    targets = collections.defaultdict(dict)
    for uid, version, fpath in [(i.split("_")[-2], int(i.split("_")[-1].split(".")[0]), i)
               for i in glob.glob(os.path.join(io_dir, "*.*"))]:
        targets[uid][version] = fpath

    # Document collection - all.
    if filter is None and not latest:
        docs = sum([i.values() for i in targets.values()], [])
        return [read(i) for i in docs]

    # Document collection - latest.
    elif filter is None and latest:
        docs = [i[max(i.keys())] for i in targets.values()]
        return [read(i) for i in docs]

    # Document collection (typed) - all.
    elif filter in ontologies.type_info.TYPES and not latest:
        docs = sum([i.values() for i in targets.values()], [])
        docs = [i for i in docs if i.find(filter.type_key) > -1]
        return [read(i) for i in docs]

    # Document collection (typed) - latest.
    elif filter in ontologies.type_info.TYPES and latest:
        docs = [i[max(i.keys())] for i in targets.values()]
        docs = [i for i in docs if i.find(filter.type_key) > -1]
        return [read(i) for i in docs]

    # Specific document - specific version.
    elif isinstance(filter, tuple) and len(filter) == 2 and isinstance(filter[1], int):
        try:
            doc = targets[filter[0]][filter[1]]
        except AttributeError:
            return None
        else:
            return read(doc)

    # Specific document - latest version.
    elif latest:
        try:
            docs = targets[filter]
        except AttributeError:
            return None
        else:
            return read(docs[max(docs.keys())])

    # Specific document - all versions.
    else:
        try:
            docs = targets[filter]
        except AttributeError:
            return None
        else:
            return [read(i) for i in docs.values()]


def convert(fpath, to_encoding, from_encoding=None):
    """Converts a document upon file system.

    :param str fpath: Path to previously saved file.
    :param str to_encoding: Target encoding.
    :param str from_encoding: Source encoding (defaults to json).

    :returns: Path to converted pyesdoc document instance.
    :rtype: str

    """
    if not os.path.isfile(fpath):
        raise IOError("Document file path does not exist")

    document = read(fpath, from_encoding)
    fpath = "{0}.{1}".format(os.path.splitext(fpath)[0], to_encoding)

    return write(document, fpath, to_encoding)





# -*- coding: utf-8 -*-

"""
.. module:: cim.v2.__init__.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The cim v2 package initialisor.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using the pyesdoc framework.

"""
# Ontology name.
NAME = 'cim'

# Ontology Version.
VERSION = '2'

# Ontology full name.
FULL_NAME = '{}.{}'.format(NAME, VERSION)

# Ontology packag set.
import typeset_for_activity_package as activity
import typeset_for_data_package as data
import typeset_for_designing_package as designing
import typeset_for_drs_package as drs
import typeset_for_platform_package as platform
import typeset_for_science_package as science
import typeset_for_shared_package as shared
import typeset_for_software_package as software
import typeset_for_time_package as time

# Ontology class set.
from typeset_for_activity_package import Activity
from typeset_for_activity_package import AxisMember
from typeset_for_activity_package import Conformance
from typeset_for_activity_package import Ensemble
from typeset_for_activity_package import EnsembleAxis
from typeset_for_activity_package import EnsembleMember
from typeset_for_activity_package import ParentSimulation
from typeset_for_activity_package import UberEnsemble
from typeset_for_data_package import Dataset
from typeset_for_data_package import Downscaling
from typeset_for_data_package import InputDataset
from typeset_for_data_package import Simulation
from typeset_for_data_package import VariableCollection
from typeset_for_designing_package import AxisMember
from typeset_for_designing_package import DomainRequirements
from typeset_for_designing_package import EnsembleRequirement
from typeset_for_designing_package import ForcingConstraint
from typeset_for_designing_package import InitialisationRequirement
from typeset_for_designing_package import MultiEnsemble
from typeset_for_designing_package import NumericalExperiment
from typeset_for_designing_package import NumericalRequirement
from typeset_for_designing_package import OutputRequirement
from typeset_for_designing_package import Project
from typeset_for_designing_package import SimulationPlan
from typeset_for_designing_package import StartDateEnsemble
from typeset_for_designing_package import TemporalConstraint
from typeset_for_drs_package import DrsAtomicDataset
from typeset_for_drs_package import DrsEnsembleIdentifier
from typeset_for_drs_package import DrsExperiment
from typeset_for_drs_package import DrsGeographicalIndicator
from typeset_for_drs_package import DrsPublicationDataset
from typeset_for_drs_package import DrsSimulationIdentifier
from typeset_for_drs_package import DrsTemporalIdentifier
from typeset_for_platform_package import ComponentPerformance
from typeset_for_platform_package import ComputePool
from typeset_for_platform_package import Machine
from typeset_for_platform_package import Partition
from typeset_for_platform_package import Performance
from typeset_for_platform_package import StoragePool
from typeset_for_platform_package import StorageVolume
from typeset_for_science_package import Model
from typeset_for_science_package import Realm
from typeset_for_science_package import Topic
from typeset_for_science_package import TopicProperty
from typeset_for_science_package import TopicPropertySet
from typeset_for_shared_package import Citation
from typeset_for_shared_package import DocMetaInfo
from typeset_for_shared_package import DocReference
from typeset_for_shared_package import ExtraAttribute
from typeset_for_shared_package import OnlineResource
from typeset_for_shared_package import Party
from typeset_for_shared_package import QualityReview
from typeset_for_shared_package import Responsibility
from typeset_for_shared_package import TextBlob
from typeset_for_software_package import ComponentBase
from typeset_for_software_package import Composition
from typeset_for_software_package import DevelopmentPath
from typeset_for_software_package import EntryPoint
from typeset_for_software_package import Gridspec
from typeset_for_software_package import Implementation
from typeset_for_software_package import SoftwareComponent
from typeset_for_software_package import Variable
from typeset_for_time_package import Calendar
from typeset_for_time_package import DateTime
from typeset_for_time_package import DatetimeSet
from typeset_for_time_package import IrregularDateset
from typeset_for_time_package import RegularTimeset
from typeset_for_time_package import TimePeriod

# Ontology enum set.
from typeset_for_activity_package import ConformanceType
from typeset_for_data_package import DataAssociationTypes
from typeset_for_designing_package import EnsembleTypes
from typeset_for_designing_package import ExperimentalRelationships
from typeset_for_designing_package import ForcingTypes
from typeset_for_designing_package import NumericalRequirementScope
from typeset_for_drs_package import DrsFrequencyTypes
from typeset_for_drs_package import DrsGeographicalOperators
from typeset_for_drs_package import DrsRealms
from typeset_for_drs_package import DrsTimeSuffixes
from typeset_for_platform_package import StorageSystems
from typeset_for_platform_package import VolumeUnits
from typeset_for_science_package import ModelTypes
from typeset_for_shared_package import NilReason
from typeset_for_shared_package import QualityStatus
from typeset_for_shared_package import RoleCode
from typeset_for_shared_package import TextBlobEncoding
from typeset_for_software_package import CouplingFramework
from typeset_for_software_package import ProgrammingLanguage
from typeset_for_time_package import CalendarTypes
from typeset_for_time_package import PeriodDateTypes
from typeset_for_time_package import TimeUnits

import type_info
import typeset
try:
	import decoder
except ImportError:
	pass



"""
.. module:: pyesdoc.mp.ontologies.schemas.cim.__init__.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Sub-package initializer.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from pyesdoc.mp.ontologies.schemas.cim import v1
from pyesdoc.mp.ontologies.schemas.cim import v2


# -*- coding: utf-8 -*-

"""
.. module:: activity_enums.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 activity package enum definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def conformance_type():
    """Creates and returns instance of conformance_type enum.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('not-xxx', None),
            ('not conformant', 'Describes a simulation that is purpefully non-conformant to an experimental requirement.'),
            ('standard config', 'Describes a simulation that is naturally conformant to an experimental requirement.'),
            ('via inputs', 'Describes a simulation that conforms to an experimental requirement by using particular inputs.'),
            ('via model mods', 'Describes a simulation that conforms to an experimental requirement by changing the configuration of the software model implementing that simulation.'),
            ('combination', 'Describes a simulation that conforms to an experimental requirement by using more than one method.'),
        ],
    }


def downscaling_type():
    """Creates and returns instance of downscaling_type enum.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('statistical', None),
            ('dynamic', None),
        ],
    }


def ensemble_type():
    """Creates and returns instance of ensemble_type enum.

    """
    return {
        'type': 'enum',
        'is_open': True,
        'members': [],
    }


def experiment_relationship_type():
    """Creates and returns instance of experiment_relationship_type enum.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('previousRealisation', None),
            ('continuationOf', None),
            ('controlExperiment', None),
            ('higherResolutionVersionOf', None),
            ('lowerResolutionVersionOf', None),
            ('increaseEnsembleOf', None),
            ('modifiedInputMethodOf', None),
            ('shorterVersionOf', None),
            ('extensionOf', None),
        ],
    }


def frequency_type():
    """Creates and returns instance of frequency_type enum.

    """
    return {
        'type': 'enum',
        'is_open': True,
        'members': [],
    }


def project_type():
    """Creates and returns instance of project_type enum.

    """
    return {
        'type': 'enum',
        'is_open': True,
        'members': [],
    }


def resolution_type():
    """Creates and returns instance of resolution_type enum.

    """
    return {
        'type': 'enum',
        'is_open': True,
        'members': [],
    }


def simulation_relationship_type():
    """Creates and returns instance of simulation_relationship_type enum.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('extensionOf', None),
            ('responseTo', None),
            ('continuationOf', None),
            ('previousSimulation', None),
            ('higherResolutionVersionOf', None),
            ('lowerResolutionVersionOf', None),
            ('fixedVersionOf', None),
            ('followingSimulation', None),
        ],
    }


def simulation_type():
    """Creates and returns instance of simulation_type enum.

    """
    return {
        'type': 'enum',
        'is_open': False,
        'members': [
            ('simulationRun', None),
            ('assimilation', None),
            ('simulationComposite', None),
        ],
    }



"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Charlotte Pascoe, David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Atmospheric chemistry startospheric heterogeneous chemistry'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Gas phase chemistry attributes',
    'properties': [
        ('gas_phase_species', 'ENUM:gas_phase_species', '0.N',
            'Gas phase species included in the stratospheric heterogeneous chemistry scheme.'),
        ('aerosol_species', 'ENUM:aerosol_species', '0.N',
            'Aerosol species included in the stratospheric heterogeneous chemistry scheme.'),
        ('number_of_reactions', 'int', '1.1',
             'The number of reactions in the stratospheric heterogeneous chemistry scheme.'),
        ('number_of_advected_species', 'int', '1.1',
             'The number of advected species in the stratospheric heterogeneous chemistry scheme.'),
        ('number_of_steady_state_species', 'int', '1.1',
             'The number of steady state species in the stratospheric heterogeneous chemistry scheme.'),
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Sedimentation
# --------------------------------------------------------------------
DETAILS['sedimentation'] = {
    'description': 'Sedimentation is the vertical settling of particles.',
    'properties': [
        ('included', 'bool', '1.1',
            'Is sedimentation is included in the stratospheric heterogeneous chemistry scheme or not?'),       
    ],
}

# --------------------------------------------------------------------
# SUB-PROCESS: Coagulation
# --------------------------------------------------------------------
DETAILS['coagulation'] = {
    'description': 'Coagulation is a process by which aerosol particles grow.',
    'properties': [
        ('included', 'bool', '1.1',
            'Is coagulation is included in the stratospheric heterogeneous chemistry scheme or not?'),       
    ],
}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['gas_phase_species'] = {
    'description': 'Gas phase species included in the stratospheric heterogeneous chemistry scheme',
    'is_open': False,
    'members': [
        ('Cly', None),
        ('Bry', None),
        ('NOy', None),
    ]
}

ENUMERATIONS['aerosol_species'] = {
    'description': 'Aerosol species included in the stratospheric heterogeneous chemistry scheme',
    'is_open': False,
    'members': [
        ('Sulphate', None),
        ('Polar stratospheric ice', None),
        ('NAT (Nitric acid trihydrate)', None),
        ('NAD (Nitric acid dihydrate)', None),
        ('STS (supercooled ternary solution aerosol particule))', None),
    ]
}



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.package
   :platform: Unix, Windows
   :synopsis: Represents an ontological package definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
from operator import add



class Package(object):
    """Represents a package within an ontology.

    :ivar name: Package name.
    :ivar doc_string: Package documentation string.
    :ivar classes: Set of associated classes.
    :ivar enums: Set of associated enums.

    """

    def __init__(self, name, doc_string, classes, enums):
        """Instance constructor.

        :param str name: Package name.
        :param str doc_string: Package documentation string.
        :param iterable classes: Set of associated classes.
        :param iterable enums: Set of associated enums.

        """
        self.abstract_classes = _get_sorted(classes, lambda c: c.is_abstract)
        self.associated = []
        self.associated_for_import = []
        self.concrete_classes = _get_sorted(classes, lambda c: not c.is_abstract)
        self.classes = _get_sorted(classes)
        self.decodings = reduce(add, [c.decodings for c in classes])
        self.doc_string = doc_string
        self.entities = []
        self.enums = _get_sorted(enums)
        self.external_types = []
        self.name = name
        self.op_name = None
        self.properties = reduce(add, [c.properties for c in classes])
        self.sub_classed = tuple()
        self.types = _get_sorted(classes + enums)


    def __repr__(self):
        """Instance string representation.

        """
        return self.name


def _get_sorted(collection, predicate=None, sort_key=None):
    """Returns a sorted collection.

    """
    if predicate:
        collection = [i for i in collection if predicate(i)]
    if sort_key is None:
        sort_key = lambda i: i.name

    return sorted(collection, key=sort_key)



# -*- coding: utf-8 -*-

"""
.. module:: dictionary.encoder.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: Enocdes a document as a python dictionary.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>

"""
from pyesdoc import ontologies



def _is_pyesdoc_type(obj):
    """Returns flag indicating whether an object is a pyesdoc type.

    """
    return isinstance(obj, ontologies.get_types())


def _is_encodable_attribute(name):
    """Returns flag indicating whether an attribute is encodable.

    """
    if name.startswith("_") or name.startswith("__") or name == "ext":
        return False
    return True


def encode(doc):
    """Encodes a document.

    :param doc: Document being encoded.
    :type doc: object

    :returns: An encoded document representation.
    :rtype: dict

    """
    obj = dict()

    for key, val in doc.__dict__.items():
        # Escape private/magic properties.
        if not _is_encodable_attribute(key):
            continue

        # Process iterables / non-iterables differently.
        try:
            iter(val)

        # Encode non-iterables:
        except TypeError:
            # ... pyesdoc types;
            if _is_pyesdoc_type(val):
                obj[key] = encode(val)
            # ... simple types;
            elif val is not None:
                obj[key] = val

        # Encode iterables:
        else:
            if len(val) > 0:
                # ... string types;
                if isinstance(val, (str, unicode)):
                    obj[key] = val
                # ... collections;
                else:
                    obj[key] = [encode(i) if _is_pyesdoc_type(i) else i for i in val]

    # Inject type info to simplify decoding.
    if doc.__class__.__name__ != "DocMetaInfo":
        if 'meta' not in obj:
            obj['meta'] = {}
        obj['meta']['type'] = type(doc).type_key

    return obj



"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Land ice glaciers'

# --------------------------------------------------------------------
# PROCESS: top level properties
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Key properties of land ice glaciers',
    'properties': [
        ('description', 'str', '1.1',
             'Describe the treatment of glaciers, if any'),
        ('dynamic_areal_extent', 'bool', '0.1',
             'Does the model include a dynamic glacial extent?'),
    ],
}

## --------------------------------------------------------------------
## SUB-PROCESS: Mass Balance
## --------------------------------------------------------------------
#DETAILS['mass_balance'] = {
#    'description': 'TODO',
#    'properties': [
#        ('calculated_in_atmosphere', 'bool', '1.1',
#             'Is the ice sheet mass balance calculated in the atmosphere/land surface components?'),
#        ('ablation_calculation', 'ENUM:ablation_calculation_types', '0.1',
#             'If mass balance is calculated in land ice model, the type of scheme used to calculate ablation in the ice sheet mass balance'),
#        ('downscaling_technique', 'str', '0.1',
#             'If mass balance is calculated in land ice model, describe how the atmospheric variables are used in the mass balance calculations'),
#    ]
#}
#
#DETAILS['snow_treatment'] = {
#    'description': 'TODO',
#    'properties': [
#        ('calculated_in_atmosphere', 'bool', '1.1',
#             'Is the glacier snow calculated in the atmosphere/land surface components?'),
#        ('number_of_snow_layers', 'int', '0.1',
#             'If glacier snow is calculated in the land ice model, how many layers in the glacier snow model?'),
#        ('processes', 'ENUM:snow_processes', '0.N',
#             'If glacier snow is calculated in the land ice model, describe processes describing snow on glaciers'),
#    ],
#}

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
#ENUMERATIONS['ablation_calculation_types'] = {
#    'description': 'Type of scheme used to model glacier ablation',
#    'is_open': True,
#    'members': [
#        ('Energy balance model', None),
#        ('PDD', None),
#    ]
#}
#
#ENUMERATIONS['glacier_snow_methods'] = {
#    'description': 'Treatment of glacier snow',
#    'is_open': True,
#    'members': [
#        ('same as snow in land surface', None),
#        ('different from snow in land surface', None),
#        ]
#    }
#
#ENUMERATIONS['snow_processes'] = {
#    'description': 'Processes describing snow on glaciers',
#    'is_open': True,
#    'members': [
#        ('prognostic snow albedo', None),
#        ('prognostic snow density', None),
#        ('prognostic snow water equivalent', None),
#        ('prognostic snow heat content', None),
#        ('prognostic snow temperature', None),
#        ('prognostic snow liquid water content', None),
#        ('snow melting', None),
#        ('snow refreezing', None),
#        ('blowing snow', None),
#    ]
#}



# -*- coding: utf-8 -*-

"""
.. module:: shared_enums.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Set of CIM v1 shared package enum definitions.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""



def change_property_type():
    """Creates and returns instance of change_property_type enum.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('InputMod', None),
            ('ModelMod', None),
            ('Decrement', None),
            ('Increment', None),
            ('Redistribution', None),
            ('Replacement', None),
            ('ParameterChange', 'A specific type of ModelMod'),
            ('CodeChange', 'A specific type of ModelMod'),
            ('AncillaryFile', 'A specific type of ModelMod'),
            ('BoundaryCondition', 'A specific type of ModelMod'),
            ('InitialCondition', 'A specific type of ModelMod'),
            ('Unused', None),
        ],
    }


def compiler_type():
    """Creates and returns instance of compiler_type enum.

    """
    return {
        'type' : 'enum',
        'is_open' : True
    }


def data_purpose():
    """Purpose of the data - i.e. ancillaryFile, boundaryCondition or initialCondition.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('ancillaryFile', None),
            ('boundaryCondition', None),
            ('initialCondition', None),
        ],
    }


def interconnect_type():
    """A list of connectors between machines.

    """
    return {
        'type' : 'enum',
        'is_open' : True
    }


def machine_type():
    """A list of known machines.

    """
    return {
        'type' : 'enum',
        'is_open' : False,
        'members' : [
            ('Parallel', None),
            ('Vector', None),
            ('Beowulf', None),
        ],
    }


def machine_vendor_type():
    """A list of organisations that create machines.

    """
    return {
        'type' : 'enum',
        'is_open' : True
    }


def operating_system_type():
    """A list of common operating systems.

    """
    return {
        'type' : 'enum',
        'is_open' : True
    }


def processor_type():
    """A list of known cpu's.

    """
    return {
        'type' : 'enum',
        'is_open' : True
    }


def unit_type():
    """A list of scientific units.

    """
    return {
        'type' : 'enum',
        'is_open' : True
    }





"""A process specialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'Eric Guilyardi'

# --------------------------------------------------------------------
# AUTHORS: Set to specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'Eric Guilyardi'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Ocean Vertical Physics'

# --------------------------------------------------------------------
# SUB-PROCESS: Boundary Layer Mixing.
# --------------------------------------------------------------------
DETAILS['boundary_layer_mixing'] = {
    'description': 'Properties of boundary layer mixing in the ocean (aka mixed layer)',
    'properties': [],
    'detail_sets': [
        'details',
        'tracers',
        'momentum'
        ]
    }

DETAILS['boundary_layer_mixing:details'] = {
    'description': 'Properties of vertical physics in ocean',
    'properties': [
        ('langmuir_cells_mixing', 'bool', '1.1',
            'Is there Langmuir cells mixing in upper ocean ?'),
        ]
    }

DETAILS['boundary_layer_mixing:tracers'] = {
    'description': 'Properties of boundary layer (BL) mixing on tracers in the ocean ',
    'properties': [
        ('type', 'ENUM:bndlayer_mixing_types', '1.1',
            'Type of boundary layer mixing for tracers in ocean'),
        ('closure_order', 'float', '0.1',
            'If turbulent BL mixing of tracers, specific order of closure (0, 1, 2.5, 3)'),
        ('constant', 'int', '0.1',
            'If constant BL mixing of tracers, specific coefficient (m2/s)'),
        ('background', 'str', '1.1',
            'Background BL mixing of tracers coefficient, (schema and value in m2/s - may by none)'),
        ]
    }

DETAILS['boundary_layer_mixing:momentum'] = {
    'description': 'Properties of boundary layer (BL) mixing on momentum in the ocean ',
    'properties': [
        ('type', 'ENUM:bndlayer_mixing_types', '1.1',
            'Type of boundary layer mixing for momentum in ocean'),
        ('closure_order', 'float', '0.1',
            'If turbulent BL mixing of momentum, specific order of closure (0, 1, 2.5, 3)'),
        ('constant', 'int', '0.1',
            'If constant BL mixing of momentum, specific coefficient (m2/s)'),
        ('background', 'str', '1.1',
            'Background BL mixing of momentum coefficient, (schema and value in m2/s - may by none)'),

        ]
    }

# --------------------------------------------------------------------
# SUB-PROCESS: Interior Layer Mixing.
# --------------------------------------------------------------------
DETAILS['interior_mixing'] = {
    'description': 'Properties of interior vertical mixing in the ocean',
    'properties': [],
    'detail_sets': [
        'details',
        'tracers',
        'momentum'
        ]
    }

DETAILS['interior_mixing:details'] = {
    'description': 'Properties of interior mixing in the ocean ',
    'properties': [
        ('convection_type', 'ENUM:vertphys_convection_types', '1.1',
            'Type of vertical convection in ocean'),
        ('tide_induced_mixing', 'str', '1.1',
            'Describe how tide induced mixing is modelled (barotropic, baroclinic, none)'),
        ('double_diffusion', 'bool', '1.1',
            'Is there double diffusion'),
        ('shear_mixing', 'bool', '1.1',
            'Is there interior shear mixing'),
        ]
    }

DETAILS['interior_mixing:tracers'] = {
    'description': 'Properties of interior mixing on tracers in the ocean ',
    'properties': [
        ('type', 'ENUM:interior_mixing_types', '1.1',
         'Type of interior mixing for tracers in ocean'),
        ('constant', 'int', '0.1',
            'If constant interior mixing of tracers, specific coefficient (m2/s)'),
        ('profile', 'str', '1.1',
            'Is the background interior mixing using a vertical profile for tracers (i.e is NOT constant) ?'),
        ('background', 'str', '1.1',
            'Background interior mixing of tracers coefficient, (schema and value in m2/s - may by none)'),
        ]
    }

DETAILS['interior_mixing:momentum'] = {
    'description': 'Properties of interior mixing on momentum in the ocean ',
    'properties': [
        ('type', 'ENUM:interior_mixing_types', '1.1',
         'Type of interior mixing for momentum in ocean'),
        ('constant', 'int', '0.1',
            'If constant interior mixing of momentum, specific coefficient (m2/s)'),
        ('profile', 'str', '1.1',
            'Is the background interior mixing using a vertical profile for momentum (i.e is NOT constant) ?'),
        ('background', 'str', '1.1',
            'Background interior mixing of momentum coefficient, (schema and value in m2/s - may by none)'),
        ]
    }

# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['vertphys_convection_types'] = {
    'description': 'Types of convection scheme in ocean',
    'is_open': True,
    'members': [
        ('Non-penetrative convective adjustment', None),
        ('Enhanced vertical diffusion', None),
        ('Included in turbulence closure', None)
        ]
    }

ENUMERATIONS['bndlayer_mixing_types'] = {
    'description': 'Types of boundary layer mixing in ocean',
    'is_open': True,
    'members': [
        ('Constant value', None),
        ('Turbulent closure - TKE', None),
        ('Turbulent closure - KPP', None),
        ('Turbulent closure - Mellor-Yamada', None),
        ('Turbulent closure - Bulk Mixed Layer', None),
        ('Richardson number dependent - PP', None),
        ('Richardson number dependent - KT', None),
        ('Imbeded as isopycnic vertical coordinate', None)
        ]
    }

ENUMERATIONS['interior_mixing_types'] = {
    'description': 'Types of interior mixing in ocean',
    'is_open': True,
    'members': [
        ('Constant value', None),
        ('Turbulent closure / TKE', None),
        ('Turbulent closure - Mellor-Yamada', None),
        ('Richardson number dependent - PP', None),
        ('Richardson number dependent - KT', None),
        ('Imbeded as isopycnic vertical coordinate', None)
        ]
    }



# -*- coding: utf-8 -*-

"""
.. module:: cim.v1.__init__.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: CIM v1 ontology schema.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""

from . import activity_classes
from . import activity_enums
from . import data_classes
from . import data_enums
from . import grids_classes
from . import grids_enums
from . import misc_classes
from . import quality_classes
from . import quality_enums
from . import shared_classes
from . import shared_classes_doc
from . import shared_classes_time
from . import shared_enums
from . import software_classes
from . import software_enums


# Ontology name.
NAME = 'cim'

# Ontology version.
VERSION = '1'

# Ontology doc string.
DOC = 'Metafor CIM ontology schema - version 1'




def activity():
    """Types that describe context against which climate models are run.

    """
    return {
        activity_classes,
        activity_enums
    }


def data():
    """Types that describe output that climate models emit.

    """
    return {
        data_classes,
        data_enums
    }


def grids():
    """Types that describe the grids that climate models plot.

    """
    return {
        grids_classes,
        grids_enums
    }


def misc():
    """Miscellaneous types.

    """
    return {
        misc_classes
    }


def quality():
    """Types that describe the quailty of output that climate models emit.

    """
    return {
        quality_classes,
        quality_enums
    }


def shared():
    """Shared types that might be imported from other packages within the ontology.

    """
    return {
        shared_classes,
        shared_classes_doc,
        shared_classes_time,
        shared_enums
    }


def software():
    """Types that describe the climate models software.

    """
    return {
        software_classes,
        software_enums
    }



# -*- coding: utf-8 -*-
"""
.. module:: cim.__init__.py

   :license: GPL / CeCILL
   :platform: Unix, Windows
   :synopsis: The cim package initializer.

.. moduleauthor:: Earth System Documentation (ES-DOC) <dev@es-doc.org>
.. note:: Code generated using pyesdoc @ 2013-04-17 13:14:13.903808.

"""
import v1
import v2



# Map of document types to display names.
DISPLAY_NAMES = {
    v1.activity.NumericalExperiment: u"Experiment",
    v1.activity.SimulationRun: u"Simulation",
    v1.data.DataObject: u"Data Object",
    v1.grids.GridSpec: u"Grid Spec",
    v1.misc.DocumentSet: u"Simulation",
    v1.quality.CimQuality: u"Dataset QC",
    v1.software.ModelComponent: u"Model",
    v2.designing.NumericalExperiment: u"Experiment",
    v2.designing.Project: u"Sub MIP"
}

# Map of document types to sort keys.
SORT_KEYS = {
    v1.activity.Ensemble: u"BC",
    v1.activity.NumericalExperiment: u"AB",
    v1.activity.SimulationRun: u"AC",
    v1.data.DataObject: u"CA",
    v1.grids.GridSpec: u"BB",
    v1.misc.DocumentSet: u"AC",
    v1.quality.CimQuality: u"CB",
    v1.shared.Platform: u"BA",
    v1.software.ModelComponent: u"AA",
    v2.designing.NumericalExperiment: u"AB",
    v2.designing.Project: u"AA"
}



# -*- coding: utf-8 -*-

"""
.. module:: security.py
   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Wraps security related functions.

.. moduleauthor:: Mark A. Conway-Greenslade


"""
import json
import os

import requests



# GitHub - user name.
_GH_USER_NAME = 'esdoc-system-user'

# GitHub - access token.
_GH_ACCESS_TOKEN = os.getenv('ESDOC_GITHUB_ACCESS_TOKEN')

# GitHub API - credentials.
_GH_API_CREDENTIALS = (_GH_USER_NAME, _GH_ACCESS_TOKEN)

# GitHub API - organization.
_GH_API_ORG = "https://api.github.com/orgs/ES-DOC-INSTITUTIONAL"

# GitHub API - teams.
_GH_API_TEAMS = "https://api.github.com/teams"

# GitHub API - user.
_GH_API_USER = "https://api.github.com/user"

# Local cache of ES-DOC teams.
_TEAMS = None


def get_team(team_id):
    """Returns set of existing GH teams.

    :param str|int team_id: Identifier of an ES-DOC GitHub team.

    :returns: ES-DOC GitHub team information.
    :rtype: dict

    """
    key = 'id' if isinstance(team_id, int) else 'name'
    for team in get_teams():
        if team[key] == team_id:
            return team

    raise ValueError('GitHub team does not exist: {}'.format(team_id))


def get_teams():
    """Returns set of ES-DOC GitHub teams.

    :returns: Collection of ES-DOC GitHub teams.
    :rtype: list

    """
    global _TEAMS

    if _TEAMS is None:
        url = '{}/teams?per_page=100'.format(_GH_API_ORG)
        r = requests.get(url, auth=_GH_API_CREDENTIALS)
        _TEAMS = json.loads(r.text)

    return _TEAMS


def is_team_member(team_id, user_id):
    """Returns flag indicating whether user is a member of a team.

    :param str|int team_id: ES-DOC GitHub team identifier.
    :param str user_id: GitHub user login.

    :returns: Flag indicating whether user is a member of team.
    :rtype: bool

    """
    team_id = team_id if isinstance(team_id, int) else get_team(team_id)['id']
    url = '{}/{}/memberships/{}'.format(_GH_API_TEAMS, team_id, user_id)
    r = requests.get(url, auth=_GH_API_CREDENTIALS)

    return r.status_code == 200


def is_authenticated_user(user_id, access_token):
    """Returns flag indicating whether user credentials are recognized by GitHub.

    :param str user_id: GitHub user login.
    :param str access_token: GitHub user access token.

    :returns: Flag indicating whether user is authenticated by GitHub.
    :rtype: bool

    """
    r = requests.get(_GH_API_USER, auth=(user_id, access_token))

    return r.status_code == 200 and \
           'read:org' in r.headers['X-OAuth-Scopes'] and \
            json.loads(r.text)['login'] == user_id



# -*- coding: utf-8 -*-

"""
.. module:: software_model_component_2.py
   :platform: Unix, Windows
   :synopsis: Extends a cim.v1.software.modelcomponent document.


"""
from pyesdoc.ontologies.cim.v1 import ComponentProperty



def _create_property(p_tree, values, name, description):
    """Sets & returns a component standard property.

    """
    def append_value(p, v):
        if v is not None:
            p.values.append(v)

    # Instantiate.
    p = ComponentProperty()
    p.short_name = p.long_name = name
    p.description = description

    # Append each value.
    if values:
        try:
            iter(values)
        except ValueError:
            for v in values:
                append_value(p, v)
        else:
            append_value(p, values)

    # Append to tree.
    p_tree.append(p)

    return p


def _set_standard_properties_descriptive(c, p_tree):
    """Sets descriptive standard properties.

    """
    _create_property(p_tree,
                     c.description,
                     'Description',
                     'High-level component description')
    _create_property(p_tree,
                     c.short_name,
                     'Short Name',
                     'Abbreviated component name')
    _create_property(p_tree,
                     c.long_name,
                     'Long Name',
                     'Full component name')


def _set_standard_properties_pi(c, p_tree):
    """Sets principal investigator standard properties.

    """
    # Set PI.
    pi = None
    for rp in c.responsible_parties:
        if rp.role and rp.role.upper() == "PI":
            pi = rp
            break

    # Set PI properties.
    if pi:
        _create_property(p_tree,
                         pi.individual_name,
                         'PI Name',
                         'PI Name')
        _create_property(p_tree,
                         pi.email,
                         'PI Email Address',
                         'PI Email Address')


def _set_standard_properties_citations(c, p_tree):
    """Sets citation standard properties.

    """
    def get_citation_title(citation):
        return citation.title.strip()

    def get_citation_url(citation):
        if citation.location is None or citation.location.strip() == "":
            return "N/A"
        else:
            return citation.location.strip()

    # Set container property.
    p = _create_property(p_tree,
                         None,
                         'Citations',
                         'Set of component citations')

    # Set citation title.
    _create_property(p.sub_properties,
                     [get_citation_title(i) for i in c.citations if i.title is not None],
                     'Title',
                     'Title')

    # Set citation location.
    _create_property(p.sub_properties,
                     [get_citation_url(i) for i in c.citations if i.title is not None],
                     'Location',
                     'Location')


def set_standard_properties(c):
    """Sets standard scientific properties.

    :param c: A model component.
    :type c: pyesdoc.ontologies.cim.v1.software.ModelComponent

    """
    # Create property group.
    p = _create_property(c.ext.standard_properties,
                         None,
                         'Standard Properties',
                         'Set of component standard properties')
    # Create sub-groups.
    for setter in [
        _set_standard_properties_descriptive,
        _set_standard_properties_pi,
        _set_standard_properties_citations
    ]:
        setter(c, p.sub_properties)


def set_scientific_properties(c):
    """Sets component scientific properties.

    :param c: A model component.
    :type c: pyesdoc.ontologies.cim.v1.software.ModelComponent

    """
    # Create property group.
    group = _create_property(c.ext.scientific_properties,
                             None,
                             'Scientific Properties',
                             'Set of component scientific properties')

    # Key properties == scientific properties.
    for cp in c.properties:
        if "Key Properties" in cp.short_name or \
           "KeyProperties" in cp.short_name:
            for cpp in cp.sub_properties:
                if cpp.short_name.lower().find("qc properties") == -1:
                    group.sub_properties.append(cpp)
            return

    # Filter out QC properties.
    for cp in c.properties:
        if cp.short_name.lower().find("qc properties") == -1:
            group.sub_properties.append(cp)


def set_qc_properties(c):
    """Sets component qc properties.

    :param c: A model component.
    :type c: pyesdoc.ontologies.cim.v1.software.ModelComponent

    """
    # Create property group.
    group = _create_property(c.ext.qc_properties,
                             None,
                             'QC Properties',
                             'Set of component quality control properties')

    # Append qc properties.
    for cp in c.properties:
        if cp.short_name.lower().find("qc properties") != -1:
            group.sub_properties = cp.sub_properties
            break



# -*- coding: utf-8 -*-

"""
.. module:: pyesdoc.mp.ontologies.core.schema_validation.schema_validator
   :platform: Unix, Windows
   :synopsis: Validates ontology schema definition.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import inspect
import re



# Regular expressions.
_RE_SCHEMA_NAME = '^[a-z_]+$'
_RE_SCHEMA_VERSION = '^[0-9\.]+$'


def _validate_package_factory(ctx, module, factory, expected_type, type_description):
    """Validates a package factory function.

    """
    try:
        instance = factory()
    except:
        err = 'Invalid {0}: {1} --> must be a no-arg callable'
        err = err.format(type_description, ctx.get_name(factory, module))
        ctx.set_error(err)
        return

    if not isinstance(instance, expected_type):
        err = 'Invalid {0}: {1} --> unexpected return type (was expecting {2})'
        err = err.format(type_description, ctx.get_name(factory, module), expected_type.__name__)
        ctx.set_error(err)
        return

    if not len(instance):
        err = 'Invalid {0}: {1} --> must not return an empty collection'
        err = err.format(type_description, ctx.get_name(factory, module))
        ctx.set_error(err)


def validate(ctx):
    """Validates schema level attributes.

    :param pyesdoc.mp.ontologies.core.schema_validation.ValidationContext: Contextutal information passed between validators.

    """
    if not inspect.ismodule(ctx.schema):
        ctx.set_error('Invalid schema --> must be python modules.')
        return

    if not hasattr(ctx.schema, 'NAME'):
        ctx.set_error('Invalid schema --> required attribute NAME is missing')

    if not hasattr(ctx.schema, 'VERSION'):
        ctx.set_error('Invalid schema --> required attribute VERSION is missing')

    if not hasattr(ctx.schema, 'DOC'):
        ctx.set_error('Invalid schema --> required attribute DOC is missing')

    if not ctx.package_factories:
        ctx.set_error('Invalid schema --> must declare at least one package')

    if ctx.report:
        return

    if ctx.schema.DOC is None or not ctx.schema.DOC.strip():
        ctx.set_error('Invalid schema --> required attribute DOC is missing')

    if not re.match(_RE_SCHEMA_NAME, ctx.schema.NAME):
        ctx.set_error('Invalid schema --> NAME must be a single word in lower case')

    if not re.match(_RE_SCHEMA_VERSION, ctx.schema.VERSION):
        ctx.set_error('Invalid schema --> VERSION must be a postive integer')

    for factory in ctx.package_factories:
        _validate_package_factory(ctx, ctx.schema, factory, set, 'package')



"""A realm process sepecialization.

For further information goto http://wordpress.es-doc.org/cmip6-model-specializations.

"""
# --------------------------------------------------------------------
# INTERNAL (do not change)
# --------------------------------------------------------------------
from collections import OrderedDict

DETAILS = OrderedDict()
ENUMERATIONS = OrderedDict()

# --------------------------------------------------------------------
# CONTACT: Set to realm specialization co-ordinator.
# --------------------------------------------------------------------
CONTACT = 'David Hassell'

# --------------------------------------------------------------------
# AUTHORS: Set to realm specialization authors (comma delimited).
# --------------------------------------------------------------------
AUTHORS = 'David Hassell'

# --------------------------------------------------------------------
# QUALITY CONTROL STATUS: Set to 'draft' or 'complete'
# --------------------------------------------------------------------
QC_STATUS = 'draft'

# --------------------------------------------------------------------
# DESCRIPTION: Short description of the specialization.
# --------------------------------------------------------------------
DESCRIPTION = 'Land surface snow'

# --------------------------------------------------------------------
# PROCESS: top level
# --------------------------------------------------------------------
DETAILS['toplevel'] = {
    'description': 'Properties of land surface snow scheme',
    'properties': [
        ('tiling', 'str', '0.1',
             'Describe the snow tiling, if any.'),        
        ('number_of_snow_layers', 'int', '1.1',
             'The number of snow levels used in the land surface scheme/model'),
        ('density', 'ENUM:snow_density_methods', '1.1',
             'Description of the treatment of snow density'),
        ('water_equivalent', 'ENUM:snow_water_equivalent_methods', '1.1',
             'Description of the treatment of the snow water equivalent'),
        ('heat_content', 'ENUM:snow_heat_content_methods', '1.1',
             'Description of the treatment of the heat content of snow'),
        ('temperature', 'ENUM:snow_temperature_methods', '1.1',
             'Description of the treatment of snow temperature'),
        ('liquid_water_content', 'ENUM:snow_liquid_water_content_methods', '1.1',
             'Description of the treatment of snow liquid water'),
        ('snow_cover_fractions', 'ENUM:snow_cover_fraction_types', '1.N',
             'Specify cover fractions used in the surface snow scheme'),
        ('processes', 'ENUM:snow_processes', '1.N',
             'Snow related processes in the land surface scheme'),
        ('prognostic_variables', 'str', '1.1',
             'List the prognostic variables of the snow scheme'),
    ]
}

DETAILS['snow_albedo'] = {
    'description': 'TODO',
    'properties' : [
        ('type', 'ENUM:snow_albedo_type', '1.1',
            'Describe the treatment of snow-covered land albedo'),
        ('functions', 'ENUM:snow_albedo_function_types', '0.N',
             'If prognostic, '),
        ]
    }
# --------------------------------------------------------------------
# ENUMERATIONS
# --------------------------------------------------------------------
ENUMERATIONS['snow_albedo_type'] = {
    'description': 'Describe the treatment of snow-covered land albedo',
    'is_open': True,
    'members': [
        ('prognostic', None),
        ('prescribed', None),        
        ('constant', None),
    ]
}

ENUMERATIONS['snow_albedo_function_types'] = {
    'description': 'Describe the dependancies on snow albedo calculations',
    'is_open': True,
    'members': [
        ('vegetation type', None),
        ('snow age', None),        
        ('snow density', None),
        ('snow grain type', None),
        ('aerosol deposition', None),
    ]
}

ENUMERATIONS['snow_density_methods'] = {
    'description': 'Treatment of snow density',
    'is_open': True,
    'members': [
        ('prognostic', None),
        ('constant', None),
    ]
}

ENUMERATIONS['snow_water_equivalent_methods'] = {
    'description': 'Treatment of snow water equivalent',
    'is_open': True,
    'members': [
        ('prognostic', None),
        ('diagnostic', None),
    ]
}

ENUMERATIONS['snow_heat_content_methods'] = {
    'description': 'Treatment of snow heat content',
    'is_open': True,
    'members': [
        ('prognostic', None),
        ('diagnostic', None),
    ]
}

ENUMERATIONS['snow_temperature_methods'] = {
    'description': 'Treatment of snow temperature',
    'is_open': True,
    'members': [
        ('prognostic', None),
        ('diagnostic', None),
    ]
}


ENUMERATIONS['snow_liquid_water_content_methods'] = {
    'description': 'Treatment of snow liquid water',
    'is_open': True,
    'members': [
        ('prognostic', None),
        ('diagnostic', None),
    ]
}

ENUMERATIONS['snow_cover_fraction_types'] = {
    'description': 'Snow cover fraction types',
    'is_open': True,
    'members': [
        ('ground snow fraction', None),
        ('vegetation snow fraction', None),
    ]
}

ENUMERATIONS['snow_processes'] = {
    'description': 'Snow-related processes',
    'is_open': True,
    'members': [
        ('snow interception', None),
        ('snow melting', None),
        ('snow freezing', None),
        ('blowing snow', None),
    ]
}




# -*- coding: utf-8 -*-
"""
.. module:: pyesdoc.ontologies.__init__.py

   :license: GPL/CeCIL
   :platform: Unix, Windows
   :synopsis: Ontologies sub-package init.

.. moduleauthor:: Mark Conway-Greenslade <momipsl@ipsl.jussieu.fr>


"""
import type_info
from utils import *
